#
# Composite model for Ohm's law in the electrode
#
import pybamm

from .base_ohm import BaseModel


class Composite(BaseModel):
    """An explicit composite leading and first order solution to solid phase
    current conservation with ohm's law. Note that the returned current density is
    only the leading order approximation.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        Either 'Negative electrode' or 'Positive electrode'
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, domain, options=None):
        super().__init__(param, domain, options=options)

    def get_coupled_variables(self, variables):
        domain = self.domain

        i_boundary_cc = variables["Current collector current density [A.m-2]"]

        # import parameters and spatial variables
        L_n = self.param.n.L
        L_p = self.param.p.L
        L_x = self.param.L_x
        x_n = pybamm.standard_spatial_vars.x_n
        x_p = pybamm.standard_spatial_vars.x_p

        tor = variables[f"X-averaged {domain} electrode transport efficiency"]
        phi_s_cn = variables["Negative current collector potential [V]"]
        T = variables[f"X-averaged {domain} electrode temperature [K]"]

        sigma_eff = self.domain_param.sigma(T) * tor
        if self._domain == "negative":
            phi_s = phi_s_cn + (i_boundary_cc / sigma_eff) * (
                x_n * (x_n - 2 * L_n) / (2 * L_n)
            )
            i_s = i_boundary_cc * (1 - x_n / L_n)

        elif self.domain == "positive":
            delta_phi_p_av = variables[
                "X-averaged positive electrode surface potential difference [V]"
            ]
            phi_e_p_av = variables["X-averaged positive electrolyte potential [V]"]

            const = (
                delta_phi_p_av
                + phi_e_p_av
                + (i_boundary_cc / sigma_eff) * (L_x - L_p / 3)
            )

            phi_s = const - (i_boundary_cc / sigma_eff) * (
                x_p + (x_p - L_x) ** 2 / (2 * L_p)
            )
            i_s = i_boundary_cc * (1 - (L_x - x_p) / L_p)

        variables.update(self._get_standard_potential_variables(phi_s))
        variables.update(self._get_standard_current_variables(i_s))

        if self.domain == "positive":
            variables.update(self._get_standard_whole_cell_variables(variables))

        return variables

    def set_boundary_conditions(self, variables):
        domain, Domain = self.domain_Domain

        phi_s = variables[f"{Domain} electrode potential [V]"]
        tor = variables[f"X-averaged {domain} electrode transport efficiency"]
        i_boundary_cc = variables["Current collector current density [A.m-2]"]
        T = variables[f"X-averaged {domain} electrode temperature [K]"]

        if self.domain == "negative":
            lbc = (pybamm.Scalar(0), "Dirichlet")
            rbc = (pybamm.Scalar(0), "Neumann")

        elif self.domain == "positive":
            lbc = (pybamm.Scalar(0), "Neumann")
            sigma_eff = self.param.p.sigma(T) * tor
            rbc = (-i_boundary_cc / sigma_eff, "Neumann")

        self.boundary_conditions[phi_s] = {"left": lbc, "right": rbc}



#
# Base class for particles
#
import pybamm


class BaseParticle(pybamm.BaseSubModel):
    """
    Base class for molar conservation in particles.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain of the model either 'Negative' or 'Positive'
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`
    phase : str, optional
        Phase of the particle (default is "primary")
    """

    def __init__(self, param, domain, options, phase="primary"):
        super().__init__(param, domain, options=options, phase=phase)
        # Read from options to see if we have a particle size distribution
        domain_options = getattr(self.options, domain)
        self.size_distribution = domain_options["particle size"] == "distribution"

    def _get_effective_diffusivity(self, c, T, current):
        domain, Domain = self.domain_Domain
        domain_param = self.domain_param
        phase_param = self.phase_param
        domain_options = getattr(self.options, domain)

        # Get diffusivity (may have empirical hysteresis)
        diffusivity_option = getattr(domain_options, self.phase)["diffusivity"]
        if diffusivity_option == "single":
            D = phase_param.D(c, T)
        elif diffusivity_option == "current sigmoid":
            k = 100
            if Domain == "Positive":
                lithiation_current = current
            elif Domain == "Negative":
                lithiation_current = -current
            m_lith = pybamm.sigmoid(0, lithiation_current, k)  # lithiation_current > 0
            m_delith = 1 - m_lith  # lithiation_current < 0
            D_lith = phase_param.D(c, T, "lithiation")
            D_delith = phase_param.D(c, T, "delithiation")
            D = m_lith * D_lith + m_delith * D_delith

        # Account for stress-induced diffusion by defining a multiplicative
        # "stress factor"
        stress_option = getattr(self.options, domain)["stress-induced diffusion"]

        if stress_option == "true":
            # Ai2019 eq [12]
            sto = c / phase_param.c_max
            Omega = pybamm.r_average(phase_param.Omega(sto, T))
            E = pybamm.r_average(phase_param.E(sto, T))
            nu = phase_param.nu
            theta_M = Omega / (self.param.R * T) * (2 * Omega * E) / (9 * (1 - nu))
            stress_factor = 1 + theta_M * (c - domain_param.c_0)
        else:
            stress_factor = 1

        return D * stress_factor

    def _get_standard_concentration_variables(
        self, c_s, c_s_xav=None, c_s_rav=None, c_s_av=None, c_s_surf=None
    ):
        """
        All particle submodels must provide the particle concentration as an argument
        to this method. Some submodels solve for quantities other than the concentration
        itself, for example the 'XAveragedPolynomialProfile' models solves for the
        x-averaged concentration. In such cases the variables being solved for (set in
        'get_fundamental_variables') must also be passed as keyword arguments. If not
        passed as keyword arguments, the various average concentrations and surface
        concentration are computed automatically from the particle concentration.
        """
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        # Get surface concentration if not provided as fundamental variable to
        # solve for
        if c_s_surf is None:
            c_s_surf = pybamm.surf(c_s)
        c_s_surf_av = pybamm.x_average(c_s_surf)

        c_scale = self.phase_param.c_max

        # Get average concentration(s) if not provided as fundamental variable to
        # solve for
        if c_s_xav is None:
            c_s_xav = pybamm.x_average(c_s)
        if c_s_rav is None:
            c_s_rav = pybamm.r_average(c_s)
        if c_s_av is None:
            c_s_av = pybamm.r_average(c_s_xav)

        variables = {
            # Dimensional concentration
            f"{Domain} {phase_name}particle concentration [mol.m-3]": c_s,
            f"X-averaged {domain} {phase_name}particle "
            "concentration [mol.m-3]": c_s_xav,
            f"R-averaged {domain} {phase_name}particle "
            "concentration [mol.m-3]": c_s_rav,
            f"Average {domain} {phase_name}particle concentration [mol.m-3]": c_s_av,
            f"{Domain} {phase_name}particle surface concentration [mol.m-3]": c_s_surf,
            f"X-averaged {domain} {phase_name}particle "
            "surface concentration [mol.m-3]": c_s_surf_av,
            f"Minimum {domain} {phase_name}particle concentration [mol.m-3]"
            "": pybamm.min(c_s),
            f"Maximum {domain} {phase_name}particle concentration [mol.m-3]"
            "": pybamm.max(c_s),
            f"Minimum {domain} {phase_name}particle "
            f"Minimum {domain} {phase_name}particle "
            "surface concentration [mol.m-3]": pybamm.min(c_s_surf),
            f"Maximum {domain} {phase_name}particle "
            "surface concentration [mol.m-3]": pybamm.max(c_s_surf),
            # Dimensionless concentration
            f"{Domain} {phase_name}particle concentration": c_s / c_scale,
            f"X-averaged {domain} {phase_name}particle concentration": c_s_xav
            / c_scale,
            f"R-averaged {domain} {phase_name}particle concentration": c_s_rav
            / c_scale,
            f"Average {domain} {phase_name}particle concentration": c_s_av / c_scale,
            f"{Domain} {phase_name}particle surface concentration": c_s_surf / c_scale,
            f"X-averaged {domain} {phase_name}particle "
            "surface concentration": c_s_surf_av / c_scale,
            f"Minimum {domain} {phase_name}particle concentration": pybamm.min(c_s)
            / c_scale,
            f"Maximum {domain} {phase_name}particle concentration": pybamm.max(c_s)
            / c_scale,
            f"Minimum {domain} {phase_name}particle "
            "surface concentration": pybamm.min(c_s_surf) / c_scale,
            f"Maximum {domain} {phase_name}particle "
            "surface concentration": pybamm.max(c_s_surf) / c_scale,
            # Stoichiometry (equivalent to dimensionless concentration)
            f"{Domain} {phase_name}particle stoichiometry": c_s / c_scale,
            f"X-averaged {domain} {phase_name}particle stoichiometry": c_s_xav
            / c_scale,
            f"R-averaged {domain} {phase_name}particle stoichiometry": c_s_rav
            / c_scale,
            f"Average {domain} {phase_name}particle stoichiometry": c_s_av / c_scale,
            f"{Domain} {phase_name}particle surface stoichiometry": c_s_surf / c_scale,
            f"X-averaged {domain} {phase_name}particle "
            "surface stoichiometry": c_s_surf_av / c_scale,
            f"Minimum {domain} {phase_name}particle stoichiometry": pybamm.min(c_s)
            / c_scale,
            f"Maximum {domain} {phase_name}particle stoichiometry": pybamm.max(c_s)
            / c_scale,
            f"Minimum {domain} {phase_name}particle "
            "surface stoichiometry": pybamm.min(c_s_surf) / c_scale,
            f"Maximum {domain} {phase_name}particle "
            "surface stoichiometry": pybamm.max(c_s_surf) / c_scale,
            # Electrode extent of lithiation
            f"{Domain} electrode extent of lithiation": c_s_rav / c_scale,
            f"X-averaged {domain} electrode extent of lithiation": c_s_av / c_scale,
        }

        return variables

    def _get_standard_flux_variables(self, N_s):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        variables = {f"{Domain} {phase_name}particle flux [mol.m-2.s-1]": N_s}

        if isinstance(N_s, pybamm.Broadcast):
            N_s_xav = pybamm.x_average(N_s)
            variables.update(
                {
                    f"X-averaged {domain} {phase_name}"
                    "particle flux [mol.m-2.s-1]": N_s_xav
                }
            )
        return variables

    def _get_distribution_variables(self, R):
        """
        Forms the particle-size distributions and mean radii given a spatial variable
        R. The domains of R will be different depending on the submodel, e.g. for the
        `SingleSizeDistribution` classes R does not have an "electrode" domain.
        """
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        R_typ = self.phase_param.R_typ  # [m]
        # Particle-size distribution (area-weighted)
        f_a_dist = self.phase_param.f_a_dist(R)  # [m-1]

        # Ensure the distribution is normalised, irrespective of discretisation
        # or user input
        f_a_dist = f_a_dist / pybamm.Integral(f_a_dist, R)  # [m-1]

        # Volume-weighted particle-size distribution
        f_v_dist = R * f_a_dist / pybamm.Integral(R * f_a_dist, R)  # [m-1]

        # Number-based particle-size distribution
        f_num_dist = (f_a_dist / R**2) / pybamm.Integral(f_a_dist / R**2, R)  # [m-1]

        # True mean radii and standard deviations, calculated from the f_a_dist that
        # was given, all have units [m]
        R_num_mean = pybamm.Integral(R * f_num_dist, R)
        R_a_mean = pybamm.Integral(R * f_a_dist, R)
        R_v_mean = pybamm.Integral(R * f_v_dist, R)
        sd_num = pybamm.sqrt(pybamm.Integral((R - R_num_mean) ** 2 * f_num_dist, R))
        sd_a = pybamm.sqrt(pybamm.Integral((R - R_a_mean) ** 2 * f_a_dist, R))
        sd_v = pybamm.sqrt(pybamm.Integral((R - R_v_mean) ** 2 * f_v_dist, R))

        # X-average the means and standard deviations to give scalars
        # (to remove the "electrode" domain, if present)
        R_num_mean = pybamm.x_average(R_num_mean)
        R_a_mean = pybamm.x_average(R_a_mean)
        R_v_mean = pybamm.x_average(R_v_mean)
        sd_num = pybamm.x_average(sd_num)
        sd_a = pybamm.x_average(sd_a)
        sd_v = pybamm.x_average(sd_v)

        # X-averaged distributions, or broadcast
        if R.domains["secondary"] == [f"{domain} electrode"]:
            f_a_dist_xav = pybamm.x_average(f_a_dist)
            f_v_dist_xav = pybamm.x_average(f_v_dist)
            f_num_dist_xav = pybamm.x_average(f_num_dist)
        else:
            f_a_dist_xav = f_a_dist
            f_v_dist_xav = f_v_dist
            f_num_dist_xav = f_num_dist

            # broadcast
            f_a_dist = pybamm.SecondaryBroadcast(f_a_dist_xav, [f"{domain} electrode"])
            f_v_dist = pybamm.SecondaryBroadcast(f_v_dist_xav, [f"{domain} electrode"])
            f_num_dist = pybamm.SecondaryBroadcast(
                f_num_dist_xav, [f"{domain} electrode"]
            )

        variables = {
            f"{Domain} {phase_name}particle sizes": R / R_typ,
            f"{Domain} {phase_name}particle sizes [m]": R,
            f"{Domain} area-weighted {phase_name}particle-size"
            " distribution [m-1]": f_a_dist,
            f"{Domain} volume-weighted {phase_name}particle-size"
            " distribution [m-1]": f_v_dist,
            f"{Domain} number-based {phase_name}particle-size"
            " distribution [m-1]": f_num_dist,
            f"{Domain} area-weighted mean particle radius [m]": R_a_mean,
            f"{Domain} volume-weighted mean particle radius [m]": R_v_mean,
            f"{Domain} number-based mean particle radius [m]": R_num_mean,
            f"{Domain} area-weighted {phase_name}particle-size"
            " standard deviation [m]": sd_a,
            f"{Domain} volume-weighted {phase_name}particle-size"
            " standard deviation [m]": sd_v,
            f"{Domain} number-based {phase_name}particle-size"
            " standard deviation [m]": sd_num,
            # X-averaged sizes and distributions
            f"X-averaged {domain} {phase_name}particle sizes [m]": pybamm.x_average(R),
            f"X-averaged {domain} area-weighted {phase_name}particle-size "
            "distribution [m-1]": f_a_dist_xav,
            f"X-averaged {domain} volume-weighted {phase_name}particle-size "
            "distribution [m-1]": f_v_dist_xav,
            f"X-averaged {domain} number-based {phase_name}particle-size "
            "distribution [m-1]": f_num_dist_xav,
        }

        return variables

    def _get_standard_concentration_distribution_variables(self, c_s):
        """
        Forms standard concentration variables that depend on particle size R given
        the fundamental concentration distribution variable c_s from the submodel.
        """
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        c_scale = self.phase_param.c_max
        # Broadcast and x-average when necessary
        if c_s.domain == [f"{domain} {phase_name}particle size"] and c_s.domains[
            "secondary"
        ] != [f"{domain} electrode"]:
            # X-avg concentration distribution
            c_s_xav_distribution = pybamm.PrimaryBroadcast(
                c_s, [f"{domain} {phase_name}particle"]
            )

            # Surface concentration distribution variables
            c_s_surf_xav_distribution = c_s
            c_s_surf_distribution = pybamm.SecondaryBroadcast(
                c_s_surf_xav_distribution, [f"{domain} electrode"]
            )

            # Concentration distribution in all domains.
            c_s_distribution = pybamm.PrimaryBroadcast(
                c_s_surf_distribution, [f"{domain} {phase_name}particle"]
            )
        elif c_s.domain == [f"{domain} {phase_name}particle"] and (
            c_s.domains["tertiary"] != [f"{domain} electrode"]
        ):
            # X-avg concentration distribution
            c_s_xav_distribution = c_s

            # Surface concentration distribution variables
            c_s_surf_xav_distribution = pybamm.surf(c_s_xav_distribution)
            c_s_surf_distribution = pybamm.SecondaryBroadcast(
                c_s_surf_xav_distribution, [f"{domain} electrode"]
            )

            # Concentration distribution in all domains.
            c_s_distribution = pybamm.TertiaryBroadcast(
                c_s_xav_distribution, [f"{domain} electrode"]
            )
        elif c_s.domain == [f"{domain} {phase_name}particle size"] and c_s.domains[
            "secondary"
        ] == [f"{domain} electrode"]:
            # Surface concentration distribution variables
            c_s_surf_distribution = c_s
            c_s_surf_xav_distribution = pybamm.x_average(c_s)

            # X-avg concentration distribution
            c_s_xav_distribution = pybamm.PrimaryBroadcast(
                c_s_surf_xav_distribution, [f"{domain} {phase_name}particle"]
            )

            # Concentration distribution in all domains
            c_s_distribution = pybamm.PrimaryBroadcast(
                c_s_surf_distribution, [f"{domain} {phase_name}particle"]
            )
        else:
            c_s_distribution = c_s

            # x-average the *tertiary* domain.
            # NOTE: not yet implemented. Make 0.5 everywhere
            c_s_xav_distribution = pybamm.FullBroadcast(
                0.5,
                [f"{domain} {phase_name}particle"],
                {
                    "secondary": f"{domain} {phase_name}particle size",
                    "tertiary": "current collector",
                },
            )

            # Surface concentration distribution variables
            c_s_surf_distribution = pybamm.surf(c_s)
            c_s_surf_xav_distribution = pybamm.x_average(c_s_surf_distribution)

        c_s_rav_distribution = pybamm.r_average(c_s_distribution)
        c_s_av_distribution = pybamm.x_average(c_s_rav_distribution)

        variables = {
            # Dimensional concentration
            f"{Domain} {phase_name}particle concentration distribution "
            "[mol.m-3]": c_s_distribution,
            f"X-averaged {domain} {phase_name}particle concentration distribution "
            "[mol.m-3]": c_s_xav_distribution,
            f"R-averaged {domain} {phase_name}particle concentration distribution "
            "[mol.m-3]": c_s_rav_distribution,
            f"Average {domain} {phase_name}particle concentration "
            "distribution [mol.m-3]": c_s_av_distribution,
            f"{Domain} {phase_name}particle surface concentration"
            " distribution [mol.m-3]": c_s_surf_distribution,
            f"X-averaged {domain} {phase_name}particle surface concentration "
            "distribution [mol.m-3]": c_s_surf_xav_distribution,
            # Dimensionless concentration
            f"{Domain} {phase_name}particle concentration "
            "distribution": c_s_distribution / c_scale,
            f"X-averaged {domain} {phase_name}particle concentration "
            "distribution": c_s_xav_distribution / c_scale,
            f"R-averaged {domain} {phase_name}particle concentration "
            "distribution": c_s_rav_distribution / c_scale,
            f"Average {domain} {phase_name}particle concentration "
            "distribution": c_s_av_distribution / c_scale,
            f"{Domain} {phase_name}particle surface concentration"
            " distribution": c_s_surf_distribution / c_scale,
            f"X-averaged {domain} {phase_name}particle surface concentration"
            " distribution": c_s_surf_xav_distribution / c_scale,
            # Stoichiometry (equivalent to dimensionless concentration)
            f"{Domain} {phase_name}particle stoichiometry "
            "distribution": c_s_distribution / c_scale,
            f"X-averaged {domain} {phase_name}particle stoichiometry "
            "distribution": c_s_xav_distribution / c_scale,
            f"R-averaged {domain} {phase_name}particle stoichiometry "
            "distribution": c_s_rav_distribution / c_scale,
            f"Average {domain} {phase_name}particle stoichiometry "
            "distribution": c_s_av_distribution / c_scale,
            f"{Domain} {phase_name}particle surface stoichiometry"
            " distribution": c_s_surf_distribution / c_scale,
            f"X-averaged {domain} {phase_name}particle surface stoichiometry"
            " distribution": c_s_surf_xav_distribution / c_scale,
            # Electrode extent of lithiation
            f"{Domain} electrode extent of lithiation": c_s_rav_distribution / c_scale,
            f"X-averaged {domain} electrode extent of lithiation": c_s_av_distribution
            / c_scale,
        }
        return variables

    def _get_standard_flux_distribution_variables(self, N_s):
        """
        Forms standard flux variables that depend on particle size R given
        the flux variable N_s from the distribution submodel.
        """
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        if [f"{domain} electrode"] in N_s.domains.values():
            # N_s depends on x

            N_s_distribution = N_s
            # x-av the *tertiary* domain
            # NOTE: not yet implemented. Fill with zeros instead
            N_s_xav_distribution = pybamm.FullBroadcast(
                0,
                [f"{domain} {phase_name}particle"],
                {
                    "secondary": f"{domain} {phase_name}particle size",
                    "tertiary": "current collector",
                },
            )
        else:
            N_s_xav_distribution = N_s
            N_s_distribution = pybamm.TertiaryBroadcast(N_s, [f"{domain} electrode"])

        variables = {
            f"X-averaged {domain} {phase_name}particle flux "
            "distribution [mol.m-2.s-1]": N_s_xav_distribution,
            f"{Domain} {phase_name}particle flux "
            "distribution [mol.m-2.s-1]": N_s_distribution,
        }

        return variables

    def _get_standard_diffusivity_variables(self, D_eff):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        variables = {
            f"{Domain} {phase_name}particle effective diffusivity [m2.s-1]": D_eff,
            f"X-averaged {domain} {phase_name}particle effective "
            "diffusivity [m2.s-1]": pybamm.x_average(D_eff),
            f"Volume-averaged {domain} {phase_name}particle effective "
            "diffusivity [m2.s-1]": pybamm.r_average(pybamm.x_average(D_eff)),
        }
        return variables

    def _get_standard_diffusivity_distribution_variables(self, D_eff):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        variables = {
            f"{Domain} {phase_name}particle effective diffusivity "
            "distribution [m2.s-1]": D_eff,
            f"X-averaged {domain} {phase_name}particle effective diffusivity "
            "distribution[m2.s-1]": pybamm.x_average(D_eff),
        }

        return variables



#
# Class for SEI growth
#
import pybamm
from .base_sei import BaseModel


class SEIGrowth(BaseModel):
    """
    Class for SEI growth.

    Most of the models are from Section 5.6.4 of :footcite:t:`Marquis2020` and
    references therein.

    The ec reaction limited model is from :footcite:t:`Yang2017`.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    reaction_loc : str
        Where the reaction happens: "x-average" (SPM, SPMe, etc),
        "full electrode" (full DFN), or "interface" (half-cell model)
    options : dict
        A dictionary of options to be passed to the model.
    phase : str, optional
        Phase of the particle (default is "primary")
    cracks : bool, optional
        Whether this is a submodel for standard SEI or SEI on cracks
    """

    def __init__(
        self, param, domain, reaction_loc, options, phase="primary", cracks=False
    ):
        super().__init__(param, domain, options=options, phase=phase, cracks=cracks)
        self.reaction_loc = reaction_loc
        SEI_option = getattr(self.options, domain)["SEI"]
        if SEI_option == "ec reaction limited":
            pybamm.citations.register("Yang2017")
        else:
            pybamm.citations.register("Marquis2020")

    def get_fundamental_variables(self):
        domain, Domain = self.domain_Domain
        Ls = []
        for pos in ["inner", "outer"]:
            scale = self.phase_param.L_sei_0
            if self.reaction_loc == "x-average":
                L_av = pybamm.Variable(
                    f"X-averaged {domain} {pos} {self.reaction_name}thickness [m]",
                    domain="current collector",
                    scale=scale,
                )
                L_av.print_name = f"L_{pos}_av"
                L = pybamm.PrimaryBroadcast(L_av, f"{domain} electrode")
            elif self.reaction_loc == "full electrode":
                L = pybamm.Variable(
                    f"{Domain} {pos} {self.reaction_name}thickness [m]",
                    domain=f"{domain} electrode",
                    auxiliary_domains={"secondary": "current collector"},
                    scale=scale,
                )
            elif self.reaction_loc == "interface":
                L = pybamm.Variable(
                    f"{Domain} {pos} {self.reaction_name}thickness [m]",
                    domain="current collector",
                    scale=scale,
                )
            L.print_name = f"L_{pos}"
            Ls.append(L)

        L_inner, L_outer = Ls

        SEI_option = getattr(getattr(self.options, domain), self.phase)["SEI"]
        if SEI_option.startswith("ec reaction limited"):
            L_inner = 0 * L_inner  # Set L_inner to zero, copying domains

        variables = self._get_standard_thickness_variables(L_inner, L_outer)

        return variables

    def get_coupled_variables(self, variables):
        phase_param = self.phase_param
        domain, Domain = self.domain_Domain
        SEI_option = getattr(getattr(self.options, domain), self.phase)["SEI"]
        T = variables[f"{Domain} electrode temperature [K]"]
        # delta_phi = phi_s - phi_e
        if self.reaction_loc == "interface":
            delta_phi = variables[
                "Lithium metal interface surface potential difference [V]"
            ]
            T = pybamm.boundary_value(T, "right")
        else:
            delta_phi = variables[
                f"{Domain} electrode surface potential difference [V]"
            ]

        # Look for current that contributes to the -IR drop
        # If we can't find the interfacial current density from the main reaction, j,
        # it's ok to fall back on the total interfacial current density, j_tot
        # This should only happen when the interface submodel is "InverseButlerVolmer"
        # in which case j = j_tot (uniform) anyway
        if f"{Domain} electrode interfacial current density [A.m-2]" in variables:
            j = variables[f"{Domain} electrode interfacial current density [A.m-2]"]
        elif self.reaction_loc == "interface":
            j = variables["Lithium metal total interfacial current density [A.m-2]"]
        else:
            j = variables[
                f"X-averaged {domain} electrode total "
                "interfacial current density [A.m-2]"
            ]

        L_sei_inner = variables[f"{Domain} inner {self.reaction_name}thickness [m]"]
        L_sei_outer = variables[f"{Domain} outer {self.reaction_name}thickness [m]"]
        L_sei = variables[f"{Domain} total {self.reaction_name}thickness [m]"]

        R_sei = phase_param.R_sei
        eta_SEI = delta_phi - phase_param.U_sei - j * L_sei * R_sei
        # Thermal prefactor for reaction, interstitial and EC models
        F_RT = self.param.F / (self.param.R * T)

        # Define alpha_SEI depending on whether it is symmetric or asymmetric. This
        # applies to "reaction limited" and "EC reaction limited"
        if SEI_option.endswith("(asymmetric)"):
            alpha_SEI = phase_param.alpha_SEI
        else:
            alpha_SEI = 0.5

        if SEI_option.startswith("reaction limited"):
            # Scott Marquis thesis (eq. 5.92)
            j_sei = -phase_param.j0_sei * pybamm.exp(-alpha_SEI * F_RT * eta_SEI)

        elif SEI_option == "electron-migration limited":
            # Scott Marquis thesis (eq. 5.94)
            eta_inner = delta_phi - phase_param.U_inner
            j_sei = (eta_inner < 0) * phase_param.kappa_inner * eta_inner / L_sei_inner
        elif SEI_option == "interstitial-diffusion limited":
            # Scott Marquis thesis (eq. 5.96)
            j_sei = -(
                phase_param.D_li * phase_param.c_li_0 * self.param.F / L_sei_outer
            ) * pybamm.exp(-F_RT * delta_phi)

        elif SEI_option == "solvent-diffusion limited":
            # Scott Marquis thesis (eq. 5.91)
            j_sei = -phase_param.D_sol * phase_param.c_sol * self.param.F / L_sei_outer

        elif SEI_option.startswith("ec reaction limited"):
            # we have a linear system for j and c
            #  c = c_0 + j * L / F / D          [1] (eq 11 in the Yang2017 paper)
            #  j = - F * c * k_exp()            [2] (eq 10 in the Yang2017 paper, factor
            #                                        of a is outside the defn of j here)
            # [1] into [2] gives (F cancels in the second terms)
            #  j = - F * c_0 * k_exp() - j * L * k_exp() / D
            # rearrange
            #  j = -F * c_0* k_exp() / (1 + L * k_exp() / D)
            #  c_ec = c_0 - L * k_exp() / D / (1 + L * k_exp() / D)
            #       = c_0 / (1 + L * k_exp() / D)
            k_exp = phase_param.k_sei * pybamm.exp(-alpha_SEI * F_RT * eta_SEI)
            L_over_D = L_sei / phase_param.D_ec
            c_0 = phase_param.c_ec_0
            j_sei = -self.param.F * c_0 * k_exp / (1 + L_over_D * k_exp)
            c_ec = c_0 / (1 + L_over_D * k_exp)

            # Get variables related to the concentration
            c_ec_av = pybamm.x_average(c_ec)

            if self.reaction == "SEI on cracks":
                name = f"{Domain} EC concentration on cracks [mol.m-3]"
            else:
                name = f"{Domain} EC surface concentration [mol.m-3]"
            variables.update({name: c_ec, f"X-averaged {name}": c_ec_av})

        if SEI_option.startswith("ec reaction limited"):
            inner_sei_proportion = 0
        else:
            inner_sei_proportion = phase_param.inner_sei_proportion

        # All SEI growth mechanisms assumed to have Arrhenius dependence
        Arrhenius = pybamm.exp(
            phase_param.E_sei / self.param.R * (1 / self.param.T_ref - 1 / T)
        )

        j_inner = inner_sei_proportion * Arrhenius * j_sei
        j_outer = (1 - inner_sei_proportion) * Arrhenius * j_sei

        variables.update(self._get_standard_concentration_variables(variables))
        variables.update(self._get_standard_reaction_variables(j_inner, j_outer))

        # Add other standard coupled variables
        variables.update(super().get_coupled_variables(variables))

        return variables

    def set_rhs(self, variables):
        phase_param = self.phase_param
        domain, Domain = self.domain_Domain

        if self.reaction_loc == "x-average":
            L_inner = variables[
                f"X-averaged {domain} inner {self.reaction_name}thickness [m]"
            ]
            L_outer = variables[
                f"X-averaged {domain} outer {self.reaction_name}thickness [m]"
            ]
            j_inner = variables[
                f"X-averaged {domain} electrode inner {self.reaction_name}"
                "interfacial current density [A.m-2]"
            ]
            j_outer = variables[
                f"X-averaged {domain} electrode outer {self.reaction_name}"
                "interfacial current density [A.m-2]"
            ]

        else:
            L_inner = variables[f"{Domain} inner {self.reaction_name}thickness [m]"]
            L_outer = variables[f"{Domain} outer {self.reaction_name}thickness [m]"]
            j_inner = variables[
                f"{Domain} electrode inner {self.reaction_name}"
                "interfacial current density [A.m-2]"
            ]
            j_outer = variables[
                f"{Domain} electrode outer {self.reaction_name}"
                "interfacial current density [A.m-2]"
            ]

        # The spreading term acts to spread out SEI along the cracks as they grow.
        # For SEI on initial surface (as opposed to cracks), it is zero.
        if self.reaction == "SEI on cracks":
            if self.reaction_loc == "x-average":
                l_cr = variables[f"X-averaged {domain} particle crack length [m]"]
                dl_cr = variables[f"X-averaged {domain} particle cracking rate [m.s-1]"]
            else:
                l_cr = variables[f"{Domain} particle crack length [m]"]
                dl_cr = variables[f"{Domain} particle cracking rate [m.s-1]"]
            spreading_outer = (
                dl_cr / l_cr * (self.phase_param.L_outer_crack_0 - L_outer)
            )
            spreading_inner = (
                dl_cr / l_cr * (self.phase_param.L_inner_crack_0 - L_inner)
            )
        else:
            spreading_outer = 0
            spreading_inner = 0

        # a * j_sei / F is the rate of consumption of li moles by SEI reaction
        # 1/z_sei converts from li moles to SEI moles (z_sei=li mol per sei mol)
        # a * j_sei / (F * z_sei) is the rate of consumption of SEI moles by SEI
        # reaction
        # V_bar / a converts from SEI moles to SEI thickness
        # V_bar * j_sei / (F * z_sei) is the rate of SEI thickness change
        dLdt_SEI_inner = (
            phase_param.V_bar_inner * j_inner / (self.param.F * phase_param.z_sei)
        )
        dLdt_SEI_outer = (
            phase_param.V_bar_outer * j_outer / (self.param.F * phase_param.z_sei)
        )

        # we have to add the spreading rate to account for cracking
        SEI_option = getattr(getattr(self.options, domain), self.phase)["SEI"]
        if SEI_option.startswith("ec reaction limited"):
            self.rhs = {L_outer: -dLdt_SEI_outer + spreading_outer}
        else:
            self.rhs = {
                L_inner: -dLdt_SEI_inner + spreading_inner,
                L_outer: -dLdt_SEI_outer + spreading_outer,
            }

    def set_initial_conditions(self, variables):
        domain, Domain = self.domain_Domain
        if self.reaction_loc == "x-average":
            L_inner = variables[
                f"X-averaged {domain} inner {self.reaction_name}thickness [m]"
            ]
            L_outer = variables[
                f"X-averaged {domain} outer {self.reaction_name}thickness [m]"
            ]
        else:
            L_inner = variables[f"{Domain} inner {self.reaction_name}thickness [m]"]
            L_outer = variables[f"{Domain} outer {self.reaction_name}thickness [m]"]

        if self.reaction == "SEI on cracks":
            L_inner_0 = self.phase_param.L_inner_crack_0
            L_outer_0 = self.phase_param.L_outer_crack_0
        else:
            L_inner_0 = self.phase_param.L_inner_0
            L_outer_0 = self.phase_param.L_outer_0
        SEI_option = getattr(getattr(self.options, domain), self.phase)["SEI"]
        if SEI_option.startswith("ec reaction limited"):
            self.initial_conditions = {L_outer: L_inner_0 + L_outer_0}
        else:
            self.initial_conditions = {L_inner: L_inner_0, L_outer: L_outer_0}



from dataclasses import dataclass
from typing import Literal, Optional, Union
import numpy as np
import pybamm


@dataclass
class ProcessedVariableTimeIntegral:
    method: Literal["discrete", "continuous"]
    initial_condition: np.ndarray
    discrete_times: Optional[np.ndarray]

    @staticmethod
    def from_pybamm_var(
        var: Union[pybamm.DiscreteTimeSum, pybamm.ExplicitTimeIntegral],
    ) -> "ProcessedVariableTimeIntegral":
        if isinstance(var, pybamm.DiscreteTimeSum):
            return ProcessedVariableTimeIntegral(
                method="discrete", initial_condition=0.0, discrete_times=var.sum_times
            )
        elif isinstance(var, pybamm.ExplicitTimeIntegral):
            return ProcessedVariableTimeIntegral(
                method="continuous",
                initial_condition=var.initial_condition.evaluate(),
                discrete_times=None,
            )
        else:
            raise ValueError("Unsupported variable type")  # pragma: no cover



import pybamm
import numpy as np


def graphite_mcmb2528_diffusivity_Dualfoil1998(sto, T):
    """
    Graphite MCMB 2528 diffusivity as a function of stoichiometry, in this case the
    diffusivity is taken to be a constant. The value is taken from Dualfoil [1].

    References
    ----------
    .. [1] http://www.cchem.berkeley.edu/jsngrp/fortran.html

    Parameters
    ----------
    sto: :class:`pybamm.Symbol`
        Electrode stoichiometry
    T: :class:`pybamm.Symbol`
        Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
        Solid diffusivity
    """

    D_ref = 3.9 * 10 ** (-14)
    E_D_s = 42770
    arrhenius = np.exp(E_D_s / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return D_ref * arrhenius


def graphite_ocp_Ramadass2004(sto):
    """
    Graphite Open-circuit Potential (OCP) as a function of the
    stoichiometry (theta?). The fit is taken from Ramadass 2004.

    References
    ----------
    .. [1] P. Ramadass, Bala Haran, Parthasarathy M. Gomadam, Ralph White, and Branko
    N. Popov. "Development of First Principles Capacity Fade Model for Li-Ion Cells."
    (2004)
    """

    u_eq = (
        0.7222
        + 0.1387 * sto
        + 0.029 * (sto**0.5)
        - 0.0172 / sto
        + 0.0019 / (sto**1.5)
        + 0.2808 * np.exp(0.9 - 15 * sto)
        - 0.7984 * np.exp(0.4465 * sto - 0.4108)
    )

    return u_eq


def graphite_electrolyte_exchange_current_density_Ramadass2004(
    c_e, c_s_surf, c_s_max, T
):
    """
    Exchange-current density for Butler-Volmer reactions between graphite and LiPF6 in
    EC:DMC.

    References
    ----------
    .. [1] P. Ramadass, Bala Haran, Parthasarathy M. Gomadam, Ralph White, and Branko
    N. Popov. "Development of First Principles Capacity Fade Model for Li-Ion Cells."
    (2004)

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    c_s_surf : :class:`pybamm.Symbol`
        Particle concentration [mol.m-3]
    c_s_max : :class:`pybamm.Symbol`
        Maximum particle concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]
    """
    m_ref = 4.854 * 10 ** (-6)  # (A/m2)(m3/mol)**1.5
    E_r = 37480
    arrhenius = np.exp(E_r / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return m_ref * arrhenius * c_e**0.5 * c_s_surf**0.5 * (c_s_max - c_s_surf) ** 0.5


def graphite_entropic_change_Moura2016(sto):
    """
    Graphite entropic change in open-circuit potential (OCP) at a temperature of
    298.15K as a function of the stoichiometry taken from Scott Moura's FastDFN code
    [1].

    References
    ----------
    .. [1] https://github.com/scott-moura/fastDFN

    Parameters
    ----------
    sto : :class:`pybamm.Symbol`
        stoichiometry of material (li-fraction)

    """
    # Original parametrization was expressed in terms of c_s_max, but we want to
    # express it in terms of stoichiometry only
    c_s_max = 24983.2619938437
    du_dT = (
        -1.5 * (120.0 / c_s_max) * np.exp(-120 * sto)
        + (0.0351 / (0.083 * c_s_max)) * ((np.cosh((sto - 0.286) / 0.083)) ** (-2))
        - (0.0045 / (0.119 * c_s_max)) * ((np.cosh((sto - 0.849) / 0.119)) ** (-2))
        - (0.035 / (0.05 * c_s_max)) * ((np.cosh((sto - 0.9233) / 0.05)) ** (-2))
        - (0.0147 / (0.034 * c_s_max)) * ((np.cosh((sto - 0.5) / 0.034)) ** (-2))
        - (0.102 / (0.142 * c_s_max)) * ((np.cosh((sto - 0.194) / 0.142)) ** (-2))
        - (0.022 / (0.0164 * c_s_max)) * ((np.cosh((sto - 0.9) / 0.0164)) ** (-2))
        - (0.011 / (0.0226 * c_s_max)) * ((np.cosh((sto - 0.124) / 0.0226)) ** (-2))
        + (0.0155 / (0.029 * c_s_max)) * ((np.cosh((sto - 0.105) / 0.029)) ** (-2))
    )

    return du_dT


def lico2_diffusivity_Ramadass2004(sto, T):
    """
    LiCo2 diffusivity as a function of stoichiometry, in this case the
    diffusivity is taken to be a constant. The value is taken from Ramadass 2004.

    References
    ----------
    .. [1] P. Ramadass, Bala Haran, Parthasarathy M. Gomadam, Ralph White, and Branko
    N. Popov. "Development of First Principles Capacity Fade Model for Li-Ion Cells."
    (2004)

    Parameters
    ----------
    sto: :class:`pybamm.Symbol`
        Electrode stoichiometry
    T: :class:`pybamm.Symbol`
        Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
        Solid diffusivity
    """
    D_ref = 1 * 10 ** (-14)
    E_D_s = 18550
    arrhenius = np.exp(E_D_s / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return D_ref * arrhenius


def lico2_ocp_Ramadass2004(sto):
    """
    Lithium Cobalt Oxide (LiCO2) Open-circuit Potential (OCP) as a a function of the
    stoichiometry. The fit is taken from Ramadass 2004. Stretch is considered the
    overhang area negative electrode / area positive electrode, in Ramadass 2002.

    References
    ----------
    .. [1] P. Ramadass, Bala Haran, Parthasarathy M. Gomadam, Ralph White, and Branko
    N. Popov. "Development of First Principles Capacity Fade Model for Li-Ion Cells."
    (2004)

    Parameters
    ----------
    sto : :class:`pybamm.Symbol`
       stoichiometry of material (li-fraction)

    """

    stretch = 1.13
    sto = stretch * sto

    u_eq = (
        -4.656
        + 88.669 * (sto**2)
        - 401.119 * (sto**4)
        + 342.909 * (sto**6)
        - 462.471 * (sto**8)
        + 433.434 * (sto**10)
    ) / (
        -1
        + 18.933 * (sto**2)
        - 79.532 * (sto**4)
        + 37.311 * (sto**6)
        - 73.083 * (sto**8)
        + 95.96 * (sto**10)
    )

    return u_eq


def lico2_electrolyte_exchange_current_density_Ramadass2004(c_e, c_s_surf, c_s_max, T):
    """
    Exchange-current density for Butler-Volmer reactions between lico2 and LiPF6 in
    EC:DMC.

    References
    ----------
    .. [1] P. Ramadass, Bala Haran, Parthasarathy M. Gomadam, Ralph White, and Branko
    N. Popov. "Development of First Principles Capacity Fade Model for Li-Ion Cells."
    (2004)

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    c_s_surf : :class:`pybamm.Symbol`
        Particle concentration [mol.m-3]
    c_s_max : :class:`pybamm.Symbol`
        Maximum particle concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]
    """
    m_ref = 2.252 * 10 ** (-6)  # (A/m2)(m3/mol)**1.5
    E_r = 39570
    arrhenius = np.exp(E_r / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return m_ref * arrhenius * c_e**0.5 * c_s_surf**0.5 * (c_s_max - c_s_surf) ** 0.5


def lico2_entropic_change_Moura2016(sto):
    """
    Lithium Cobalt Oxide (LiCO2) entropic change in open-circuit potential (OCP) at
    a temperature of 298.15K as a function of the stoichiometry. The fit is taken
    from Scott Moura's FastDFN code [1].

    References
    ----------
    .. [1] https://github.com/scott-moura/fastDFN

    Parameters
    ----------
    sto : :class:`pybamm.Symbol`
        stoichiometry of material (li-fraction)
    """
    # Since the equation for LiCo2 from this ref. has the stretch factor,
    # should this too? If not, the "bumps" in the OCV don't line up.
    stretch = 1.062
    sto = stretch * sto
    # Original parametrization was expressed in terms of c_s_max, but we want to
    # express it in terms of stoichiometry only
    c_s_max = 51217.9257309275
    du_dT = (
        0.07645 * (-54.4806 / c_s_max) * ((1.0 / np.cosh(30.834 - 54.4806 * sto)) ** 2)
        + 2.1581 * (-50.294 / c_s_max) * ((np.cosh(52.294 - 50.294 * sto)) ** (-2))
        + 0.14169 * (19.854 / c_s_max) * ((np.cosh(11.0923 - 19.8543 * sto)) ** (-2))
        - 0.2051 * (5.4888 / c_s_max) * ((np.cosh(1.4684 - 5.4888 * sto)) ** (-2))
        - (0.2531 / 0.1316 / c_s_max) * ((np.cosh((-sto + 0.56478) / 0.1316)) ** (-2))
        - (0.02167 / 0.006 / c_s_max) * ((np.cosh((sto - 0.525) / 0.006)) ** (-2))
    )

    return du_dT


def electrolyte_diffusivity_Ramadass2004(c_e, T):
    """
    Diffusivity of LiPF6 in EC:DMC as a function of ion concentration.

    References
    ----------
    .. [1] P. Ramadass, Bala Haran, Parthasarathy M. Gomadam, Ralph White, and Branko
    N. Popov. "Development of First Principles Capacity Fade Model for Li-Ion Cells."
    (2004)

    Parameters
    ----------
    c_e: :class:`pybamm.Symbol`
        Dimensional electrolyte concentration
    T: :class:`pybamm.Symbol`
        Dimensional temperature


    Returns
    -------
    :class:`pybamm.Symbol`
        Solid diffusivity
    """

    D_c_e = 7.5e-10
    E_D_e = 37040
    arrhenius = np.exp(E_D_e / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return D_c_e * arrhenius


def electrolyte_conductivity_Ramadass2004(c_e, T):
    """
    Conductivity of LiPF6 in EC:DMC as a function of ion concentration.
    Concentration should be in dm3 in the function.

    References
    ----------
    .. [1] P. Ramadass, Bala Haran, Parthasarathy M. Gomadam, Ralph White, and Branko
    N. Popov. "Development of First Principles Capacity Fade Model for Li-Ion Cells."
    (2004)

    Parameters
    ----------
    c_e: :class:`pybamm.Symbol`
        Dimensional electrolyte concentration
    T: :class:`pybamm.Symbol`
        Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
        Solid diffusivity
    """
    # mol.m-3 to mol.dm-3, original function is likely in mS/cm
    # The function is not in Arora 2000 as reported in Ramadass 2004

    cm = 1e-6 * c_e  # here it should be only 1e-3

    sigma_e = (
        4.1253 * (10 ** (-4))
        + 5.007 * cm
        - 4.7212 * (10**3) * (cm**2)
        + 1.5094 * (10**6) * (cm**3)
        - 1.6018 * (10**8) * (cm**4)
    ) * 1e3  # and here there should not be an exponent

    E_k_e = 34700
    arrhenius = np.exp(E_k_e / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return sigma_e * arrhenius


# Call dict via a function to avoid errors when editing in place
def get_parameter_values():
    """
    Ramadass2004 parameter set. This is a bit of a Frankenstein parameter set and should
    be used with caution.

    Parameters for a graphite negative electrode, Lithium Cobalt Oxide positive
    electrode, and LiPF6 electrolyte are from the papers :footcite:t:`Marquis2019`,
    :footcite:t:`Ramadass2004`, and references therein.

    Parameters for the separator are from the papers :footcite:t:`Ecker2015i`

    The thermal material properties are for a 5 Ah power pouch cell by Kokam. The data
    are extracted from Lithium Cobalt Oxide positive electrode parameters in
    :footcite:t:`Zhao2018`

    Parameters for SEI growth are from the papers :footcite:t:`Ramadass2004` and
    :footcite:t:`Safari2008`

    .. note::
        Ramadass 2004 has mistakes in units and values of SEI parameters, corrected by
        Safari2009.
    """

    return {
        "chemistry": "lithium_ion",
        # sei
        "Ratio of lithium moles to SEI moles": 2.0,
        "Inner SEI reaction proportion": 0.5,
        "Inner SEI partial molar volume [m3.mol-1]": 9.585e-05,
        "Outer SEI partial molar volume [m3.mol-1]": 9.585e-05,
        "SEI reaction exchange current density [A.m-2]": 1.5e-06,
        "SEI resistivity [Ohm.m]": 200000.0,
        "Outer SEI solvent diffusivity [m2.s-1]": 2.5000000000000002e-22,
        "Bulk solvent concentration [mol.m-3]": 2636.0,
        "Inner SEI open-circuit potential [V]": 0.1,
        "Outer SEI open-circuit potential [V]": 0.8,
        "Inner SEI electron conductivity [S.m-1]": 8.95e-14,
        "Inner SEI lithium interstitial diffusivity [m2.s-1]": 1e-20,
        "Lithium interstitial reference concentration [mol.m-3]": 15.0,
        "Initial inner SEI thickness [m]": 2.5e-09,
        "Initial outer SEI thickness [m]": 2.5e-09,
        "EC initial concentration in electrolyte [mol.m-3]": 4541.0,
        "EC diffusivity [m2.s-1]": 2e-18,
        "SEI kinetic rate constant [m.s-1]": 1e-12,
        "SEI open-circuit potential [V]": 0.0,
        "SEI growth activation energy [J.mol-1]": 0.0,
        # cell
        "Negative current collector thickness [m]": 1.7e-05,
        "Negative electrode thickness [m]": 8.8e-05,
        "Separator thickness [m]": 2.5e-05,
        "Positive electrode thickness [m]": 8e-05,
        "Positive current collector thickness [m]": 2.3e-05,
        "Electrode height [m]": 0.057,
        "Electrode width [m]": 1.060692,
        "Negative tab width [m]": 0.04,
        "Negative tab centre y-coordinate [m]": 0.06,
        "Negative tab centre z-coordinate [m]": 0.137,
        "Positive tab width [m]": 0.04,
        "Positive tab centre y-coordinate [m]": 0.147,
        "Positive tab centre z-coordinate [m]": 0.137,
        "Negative current collector conductivity [S.m-1]": 59600000.0,
        "Positive current collector conductivity [S.m-1]": 35500000.0,
        "Negative current collector density [kg.m-3]": 8954.0,
        "Positive current collector density [kg.m-3]": 2707.0,
        "Negative current collector specific heat capacity [J.kg-1.K-1]": 385.0,
        "Positive current collector specific heat capacity [J.kg-1.K-1]": 897.0,
        "Negative current collector thermal conductivity [W.m-1.K-1]": 401.0,
        "Positive current collector thermal conductivity [W.m-1.K-1]": 237.0,
        "Nominal cell capacity [A.h]": 1.0,
        "Current function [A]": 1.0,
        "Contact resistance [Ohm]": 0,
        # negative electrode
        "Negative electrode conductivity [S.m-1]": 100.0,
        "Maximum concentration in negative electrode [mol.m-3]": 30555.0,
        "Negative particle diffusivity [m2.s-1]"
        "": graphite_mcmb2528_diffusivity_Dualfoil1998,
        "Negative electrode OCP [V]": graphite_ocp_Ramadass2004,
        "Negative electrode porosity": 0.485,
        "Negative electrode active material volume fraction": 0.49,
        "Negative particle radius [m]": 2e-06,
        "Negative electrode Bruggeman coefficient (electrolyte)": 4.0,
        "Negative electrode Bruggeman coefficient (electrode)": 4.0,
        "Negative electrode charge transfer coefficient": 0.5,
        "Negative electrode double-layer capacity [F.m-2]": 0.2,
        "Negative electrode exchange-current density [A.m-2]"
        "": graphite_electrolyte_exchange_current_density_Ramadass2004,
        "Negative electrode density [kg.m-3]": 1657.0,
        "Negative electrode specific heat capacity [J.kg-1.K-1]": 700.0,
        "Negative electrode thermal conductivity [W.m-1.K-1]": 1.7,
        "Negative electrode OCP entropic change [V.K-1]"
        "": graphite_entropic_change_Moura2016,
        # positive electrode
        "Positive electrode conductivity [S.m-1]": 100.0,
        "Maximum concentration in positive electrode [mol.m-3]": 51555.0,
        "Positive particle diffusivity [m2.s-1]": lico2_diffusivity_Ramadass2004,
        "Positive electrode OCP [V]": lico2_ocp_Ramadass2004,
        "Positive electrode porosity": 0.385,
        "Positive electrode active material volume fraction": 0.59,
        "Positive particle radius [m]": 2e-06,
        "Positive electrode Bruggeman coefficient (electrolyte)": 4.0,
        "Positive electrode Bruggeman coefficient (electrode)": 4.0,
        "Positive electrode charge transfer coefficient": 0.5,
        "Positive electrode double-layer capacity [F.m-2]": 0.2,
        "Positive electrode exchange-current density [A.m-2]"
        "": lico2_electrolyte_exchange_current_density_Ramadass2004,
        "Positive electrode density [kg.m-3]": 3262.0,
        "Positive electrode specific heat capacity [J.kg-1.K-1]": 700.0,
        "Positive electrode thermal conductivity [W.m-1.K-1]": 2.1,
        "Positive electrode OCP entropic change [V.K-1]"
        "": lico2_entropic_change_Moura2016,
        # separator
        "Separator porosity": 0.508,
        "Separator Bruggeman coefficient (electrolyte)": 1.9804586773134945,
        "Separator density [kg.m-3]": 1017.0,
        "Separator specific heat capacity [J.kg-1.K-1]": 1978.0,
        "Separator thermal conductivity [W.m-1.K-1]": 0.34,
        # electrolyte
        "Initial concentration in electrolyte [mol.m-3]": 1000.0,
        "Cation transference number": 0.363,
        "Thermodynamic factor": 1.0,
        "Electrolyte diffusivity [m2.s-1]": electrolyte_diffusivity_Ramadass2004,
        "Electrolyte conductivity [S.m-1]": electrolyte_conductivity_Ramadass2004,
        # experiment
        "Reference temperature [K]": 298.15,
        "Ambient temperature [K]": 298.15,
        "Negative current collector surface heat transfer coefficient [W.m-2.K-1]"
        "": 0.0,
        "Positive current collector surface heat transfer coefficient [W.m-2.K-1]"
        "": 0.0,
        "Negative tab heat transfer coefficient [W.m-2.K-1]": 10.0,
        "Positive tab heat transfer coefficient [W.m-2.K-1]": 10.0,
        "Edge heat transfer coefficient [W.m-2.K-1]": 0.3,
        "Number of electrodes connected in parallel to make a cell": 1.0,
        "Number of cells connected in series to make a battery": 1.0,
        "Lower voltage cut-off [V]": 2.8,
        "Upper voltage cut-off [V]": 4.2,
        "Open-circuit voltage at 0% SOC [V]": 2.8,
        "Open-circuit voltage at 100% SOC [V]": 4.2,
        "Initial concentration in negative electrode [mol.m-3]": 22610.7,
        "Initial concentration in positive electrode [mol.m-3]": 25777.5,
        "Initial temperature [K]": 298.15,
        # citations
        "citations": ["Ramadass2004"],
    }



#
# Lead-acid LOQS model
#
import pybamm
from .base_lead_acid_model import BaseModel


class LOQS(BaseModel):
    """
    Leading-Order Quasi-Static model for lead-acid, from
    :footcite:t:`Sulzer2019asymptotic`.
    See :class:`pybamm.lead_acid.BaseModel` for more details.

    """

    def __init__(self, options=None, name="LOQS model", build=True):
        super().__init__(options, name)

        self.set_external_circuit_submodel()
        self.set_open_circuit_potential_submodel()
        self.set_intercalation_kinetics_submodel()
        self.set_interface_utilisation_submodel()
        self.set_convection_submodel()
        self.set_porosity_submodel()
        self.set_active_material_submodel()
        self.set_transport_efficiency_submodels()
        self.set_electrolyte_submodel()
        self.set_electrode_submodels()
        self.set_thermal_submodel()
        self.set_surface_temperature_submodel()
        self.set_side_reaction_submodels()
        self.set_current_collector_submodel()
        self.set_sei_submodel()
        self.set_lithium_plating_submodel()
        self.set_total_interface_submodel()

        if build:
            self.build_model()

        if self.options["dimensionality"] == 0:
            self.use_jacobian = False

        pybamm.citations.register("Sulzer2019asymptotic")

    def set_external_circuit_submodel(self):
        """
        Define how the external circuit defines the boundary conditions for the model,
        e.g. (not necessarily constant-) current, voltage, etc
        """
        if self.options["operating mode"] == "current":
            self.submodels["external circuit"] = (
                pybamm.external_circuit.ExplicitCurrentControl(self.param, self.options)
            )
        elif self.options["operating mode"] == "voltage":
            self.submodels["external circuit"] = (
                pybamm.external_circuit.VoltageFunctionControl(self.param, self.options)
            )
        elif self.options["operating mode"] == "power":
            self.submodels["external circuit"] = (
                pybamm.external_circuit.PowerFunctionControl(self.param, self.options)
            )
        elif callable(self.options["operating mode"]):
            self.submodels["external circuit"] = (
                pybamm.external_circuit.FunctionControl(
                    self.param, self.options["operating mode"], self.options
                )
            )

    def set_current_collector_submodel(self):
        if self.options["current collector"] in [
            "uniform",
            "potential pair quite conductive",
        ]:
            submodel = pybamm.current_collector.Uniform(self.param)
        elif self.options["current collector"] == "potential pair":
            if self.options["dimensionality"] == 1:
                submodel = pybamm.current_collector.PotentialPair1plus1D(self.param)
            elif self.options["dimensionality"] == 2:
                submodel = pybamm.current_collector.PotentialPair2plus1D(self.param)
        self.submodels["leading-order current collector"] = submodel

    def set_porosity_submodel(self):
        self.submodels["leading-order porosity"] = pybamm.porosity.ReactionDrivenODE(
            self.param, self.options, True
        )

    def set_convection_submodel(self):
        if self.options["convection"] == "none":
            self.submodels["leading-order transverse convection"] = (
                pybamm.convection.transverse.NoConvection(self.param)
            )
            self.submodels["leading-order through-cell convection"] = (
                pybamm.convection.through_cell.NoConvection(self.param)
            )
        else:
            if self.options["convection"] == "uniform transverse":
                self.submodels["leading-order transverse convection"] = (
                    pybamm.convection.transverse.Uniform(self.param)
                )
            elif self.options["convection"] == "full transverse":
                self.submodels["leading-order transverse convection"] = (
                    pybamm.convection.transverse.Full(self.param)
                )
            self.submodels["leading-order through-cell convection"] = (
                pybamm.convection.through_cell.Explicit(self.param)
            )

    def set_intercalation_kinetics_submodel(self):
        if self.options["surface form"] == "false":
            self.submodels["leading-order negative interface"] = (
                pybamm.kinetics.InverseButlerVolmer(
                    self.param, "negative", "lead-acid main", self.options
                )
            )
            self.submodels["leading-order positive interface"] = (
                pybamm.kinetics.InverseButlerVolmer(
                    self.param, "positive", "lead-acid main", self.options
                )
            )
            self.submodels["negative interface current"] = (
                pybamm.kinetics.CurrentForInverseButlerVolmer(
                    self.param, "negative", "lead-acid main"
                )
            )
            self.submodels["positive interface current"] = (
                pybamm.kinetics.CurrentForInverseButlerVolmer(
                    self.param, "positive", "lead-acid main"
                )
            )
        else:
            self.submodels["leading-order negative interface"] = (
                pybamm.kinetics.SymmetricButlerVolmer(
                    self.param, "negative", "lead-acid main", self.options, "primary"
                )
            )

            self.submodels["leading-order positive interface"] = (
                pybamm.kinetics.SymmetricButlerVolmer(
                    self.param, "positive", "lead-acid main", self.options, "primary"
                )
            )
        # always use forward Butler-Volmer for the reaction submodel to be passed to the
        # higher order model
        self.reaction_submodels = {
            "negative": [
                pybamm.kinetics.SymmetricButlerVolmer(
                    self.param, "negative", "lead-acid main", self.options, "primary"
                )
            ],
            "positive": [
                pybamm.kinetics.SymmetricButlerVolmer(
                    self.param, "positive", "lead-acid main", self.options, "primary"
                )
            ],
        }

    def set_electrode_submodels(self):
        self.submodels["leading-order negative electrode potential"] = (
            pybamm.electrode.ohm.LeadingOrder(self.param, "negative")
        )
        self.submodels["leading-order positive electrode potential"] = (
            pybamm.electrode.ohm.LeadingOrder(self.param, "positive")
        )

    def set_electrolyte_submodel(self):
        surf_form = pybamm.electrolyte_conductivity.surface_potential_form

        if self.options["surface form"] == "false":
            self.submodels["leading-order electrolyte conductivity"] = (
                pybamm.electrolyte_conductivity.LeadingOrder(self.param)
            )
            surf_model = surf_form.Explicit
        elif self.options["surface form"] == "differential":
            surf_model = surf_form.LeadingOrderDifferential
        elif self.options["surface form"] == "algebraic":
            surf_model = surf_form.LeadingOrderAlgebraic

        for domain in ["negative", "positive"]:
            self.submodels[f"{domain} surface potential difference"] = surf_model(
                self.param, domain, self.options
            )

        self.submodels["electrolyte diffusion"] = (
            pybamm.electrolyte_diffusion.LeadingOrder(self.param)
        )

    def set_side_reaction_submodels(self):
        if self.options["hydrolysis"] == "true":
            self.submodels["leading-order oxygen diffusion"] = (
                pybamm.oxygen_diffusion.LeadingOrder(self.param)
            )
            self.submodels["leading-order positive oxygen interface"] = (
                pybamm.kinetics.ForwardTafel(
                    self.param, "positive", "lead-acid oxygen", self.options, "primary"
                )
            )
            self.submodels["leading-order negative oxygen interface"] = (
                pybamm.kinetics.DiffusionLimited(
                    self.param,
                    "negative",
                    "lead-acid oxygen",
                    self.options,
                    order="leading",
                )
            )
        else:
            self.submodels["leading-order oxygen diffusion"] = (
                pybamm.oxygen_diffusion.NoOxygen(self.param)
            )
            self.submodels["leading-order negative oxygen interface"] = (
                pybamm.kinetics.NoReaction(
                    self.param, "negative", "lead-acid oxygen", "primary"
                )
            )
            self.submodels["leading-order positive oxygen interface"] = (
                pybamm.kinetics.NoReaction(
                    self.param, "positive", "lead-acid oxygen", "primary"
                )
            )
        self.reaction_submodels["negative"].append(
            self.submodels["leading-order negative oxygen interface"]
        )
        self.reaction_submodels["positive"].append(
            self.submodels["leading-order positive oxygen interface"]
        )



import pybamm

import numpy as np

from scipy.sparse import diags, eye, kron, csr_matrix, lil_matrix, coo_matrix, vstack


class SpectralVolume(pybamm.FiniteVolume):
    """
    A class which implements the steps specific to the Spectral Volume
    discretisation. It is implemented in such a way that it is very
    similar to FiniteVolume; that comes at the cost that it is only
    compatible with the SpectralVolume1DSubMesh (which is a certain
    subdivision of any 1D mesh, so it shouldn't be a problem).

    For broadcast and mass_matrix, we follow the default behaviour from
    SpatialMethod. For spatial_variable,
    divergence, divergence_matrix, laplacian, integral,
    definite_integral_matrix, indefinite_integral,
    indefinite_integral_matrix, indefinite_integral_matrix_nodes,
    indefinite_integral_matrix_edges, delta_function
    we follow the behaviour from FiniteVolume. This is possible since
    the node values are integral averages with Spectral Volume, just
    as with Finite Volume. delta_function assigns the integral value
    to a CV instead of a SV this way, but that doesn't matter too much.
    Additional methods that are inherited by FiniteVolume which
    technically are not suitable for Spectral Volume are
    boundary_value_or_flux, process_binary_operators, concatenation,
    node_to_edge, edge_to_node and shift. While node_to_edge (as well as
    boundary_value_or_flux and process_binary_operators)
    could utilize the reconstruction approach of Spectral Volume, the
    inverse edge_to_node would still have to fall back to the Finite
    Volume behaviour. So these are simply inherited for consistency.
    boundary_value_or_flux might not benefit from the reconstruction
    approach at all, as it seems to only preprocess symbols.

    Parameters
    ----------
    mesh : :class:`pybamm.Mesh`
        Contains all the submeshes for discretisation
    """

    def __init__(self, options=None, order=2):
        self.order = order
        super().__init__(options)
        pybamm.citations.register("Wang2002")

    def chebyshev_collocation_points(self, noe, a=-1.0, b=1.0):
        """
        Calculates Chebyshev collocation points in descending order.

        Parameters
        ----------
        noe: integer
            The number of the collocation points. "number of edges"
        a: float
            Left end of the interval on which the Chebyshev collocation
            points are constructed. Default is -1.
        b: float
            Right end of the interval on which the Chebyshev collocation
            points are constructed. Default is 1.

        Returns
        -------
        :class:`numpy.array`
        Chebyshev collocation points on [a,b].
        """

        return a + 0.5 * (b - a) * (
            1
            + np.sin(
                np.pi
                * np.array([(noe - 1 - 2 * i) / (2 * noe - 2) for i in range(noe)])
            )
        )

    def cv_boundary_reconstruction_sub_matrix(self):
        """
        Coefficients for reconstruction of a function through averages.
        The resulting matrix is scale-invariant :footcite:t:`Wang2002`.

        Parameters
        ----------

        Returns
        -------

        """

        # While Spectral Volume in general may use any point
        # distribution for CVs, the Chebyshev nodes are the most stable.
        # The differentiation matrices are only implemented for those.
        edges = np.flip(self.chebyshev_collocation_points(self.order + 1))

        # Nomenclature in the reference:
        # c[j,l] are the coefficients from the reference.
        # The index of the CV boundaries j ranges from 0 to self.order.
        # The index of the CVs themselves l ranges from 1 to self.order.
        # l ranges from 0 to self.order - 1 here.
        c = np.empty([self.order + 1, self.order])
        # h[l] are the lengths of the CVs.
        h = [edges[i + 1] - edges[i] for i in range(self.order)]

        # Optimised derivative of the "Lagrange polynomial denominator".
        # It is equivalent to d_omega_d_x(x) at x = x_{j+1/2}.
        def d_omega_d_x(j):
            return np.prod(
                edges[j] - edges,
                where=[True] * j + [False] + [True] * (len(edges) - 1 - j),
            )

        for j in range(self.order + 1):
            for ell in range(self.order):
                c[j, ell] = h[ell] * np.sum(
                    [
                        1.0
                        / d_omega_d_x(r)
                        * np.sum(
                            [
                                np.prod(
                                    edges[j] - edges,
                                    where=[
                                        q != r and q != m for q in range(self.order + 1)
                                    ],
                                )
                                for m in range(self.order + 1)
                            ],
                            where=[m != r for m in range(self.order + 1)],
                        )
                        for r in range(ell + 1, self.order + 1)
                    ]
                )

        return c

    def cv_boundary_reconstruction_matrix(self, domains):
        """
        "Broadcasts" the basic edge value reconstruction matrix to the
        actual shape of the discretised symbols. Note that the product
        of this and a discretised symbol is a vector which represents
        duplicate values for all inner SV edges. These are the
        reconstructed values from both sides.

        Parameters
        ----------
        domains : dict
            The domains in which to compute the gradient matrix

        Returns
        -------
        :class:`pybamm.Matrix`
            The (sparse) CV reconstruction matrix for the domain
        """
        submesh = self.mesh[domains["primary"]]

        # Obtain the basic reconstruction matrix.
        recon_sub_matrix = self.cv_boundary_reconstruction_sub_matrix()

        # Create 1D matrix using submesh
        # n is the number of SVs, submesh.npts is the number of CVs
        n = submesh.npts // self.order
        sub_matrix = csr_matrix(kron(eye(n), recon_sub_matrix))

        # number of repeats
        second_dim_repeats = self._get_auxiliary_domain_repeats(domains)

        # generate full matrix from the submatrix
        # Convert to csr_matrix so that we can take the index
        # (row-slicing), which is not supported by the default kron
        # format. Note that this makes column-slicing inefficient,
        # but this should not be an issue.
        matrix = csr_matrix(kron(eye(second_dim_repeats), sub_matrix))

        return pybamm.Matrix(matrix)

    def chebyshev_differentiation_matrices(self, noe, dod):
        """
        Chebyshev differentiation matrices, from
        :footcite:t:`Baltensperger2003`.

        Parameters
        ----------
        noe: integer
            The number of the collocation points. "number of edges"
        dod: integer
            The maximum order of differentiation for which a
            differentiation matrix shall be calculated. Note that it has
            to be smaller than 'noe'. "degrees of differentiation"

        Returns
        -------
        list(:class:`numpy.array`)
            The differentiation matrices in ascending order of
            differentiation order. With exact arithmetic, the diff.
            matrix of order p would just be the pth matrix power of
            the diff. matrix of order 1. This method computes the higher
            orders in a more numerically stable way.
        """
        if dod >= noe:
            raise ValueError(
                "Too many degrees of differentiation. At most "
                + str(noe - 1)
                + " are possible for "
                + str(noe)
                + " edges."
            )

        edges = self.chebyshev_collocation_points(noe)

        # These matrices tend to be dense, thus numpy arrays are used.
        prefactors = np.array(
            [[(i - j + 1) % 2 - (i - j) % 2 for j in range(noe)] for i in range(noe)]
        )
        prefactors = (prefactors * np.array([2] + [1 for i in range(noe - 2)] + [2])).T
        prefactors = prefactors * np.array([0.5] + [1 for i in range(noe - 2)] + [0.5])

        inverse_difference = np.array(
            [
                [1.0 / (edges[i] - edges[j]) for j in range(i)]
                + [0.0]
                + [1.0 / (edges[i] - edges[j]) for j in range(i + 1, noe)]
                for i in range(noe)
            ]
        )

        differentiation_matrices = []
        # This matrix changes in each of the following iterations.
        temp_diff = np.eye(noe)

        # The calculation here makes extensive use of the element-wise
        # multiplication of numpy.arrays. The * are intentionally not @!
        for p in range(dod):
            temp = (prefactors.T * np.diag(temp_diff)).T - temp_diff
            temp_diff = (p + 1) * inverse_difference * temp
            # Negative sum trick: the rows of the exact matrices sum to
            # zero. The diagonal gets less accurate with this, but the
            # approximation of the differential will be better overall.
            for i in range(noe):
                temp_diff[i, i] = -np.sum(np.delete(temp_diff[i], i))
            differentiation_matrices.append(temp_diff.copy())

        return differentiation_matrices

    def gradient(self, symbol, discretised_symbol, boundary_conditions):
        """Matrix-vector multiplication to implement the gradient
        operator. See :meth:`pybamm.SpatialMethod.gradient`
        """
        # Reconstruct edge values from node values.
        reconstructed_symbol = (
            self.cv_boundary_reconstruction_matrix(symbol.domains) @ discretised_symbol
        )

        # Add Dirichlet boundary conditions, if defined
        if symbol in boundary_conditions:
            bcs = boundary_conditions[symbol]
            if any(bc[1] == "Dirichlet" for bc in bcs.values()):
                # add ghost nodes and update domain
                reconstructed_symbol = self.replace_dirichlet_values(
                    symbol, reconstructed_symbol, bcs
                )

        # note in 1D cartesian, cylindrical and spherical grad are the same
        gradient_matrix = self.gradient_matrix(symbol.domain, symbol.domains)
        penalty_matrix = self.penalty_matrix(symbol.domains)

        # Multiply by gradient matrix
        out = (
            gradient_matrix @ reconstructed_symbol + penalty_matrix @ discretised_symbol
        )

        # Add Neumann boundary conditions, if defined
        if symbol in boundary_conditions:
            bcs = boundary_conditions[symbol]
            if any(bc[1] == "Neumann" for bc in bcs.values()):
                out = self.replace_neumann_values(symbol, out, bcs)

        return out

    def gradient_matrix(self, domain, domains):
        """
        Gradient matrix for Spectral Volume in the appropriate domain.
        Note that it contains the averaging of the duplicate SV edge
        gradient values, such that the product of it and a reconstructed
        discretised symbol simply represents CV edge values.
        On its own, it only works on non-concatenated domains, since
        only then the boundary conditions ensure correct behaviour.
        More generally, it only works if gradients are a result of
        boundary conditions rather than continuity conditions.
        For example, two adjacent SVs with gradient zero in each of them
        but with different variable values will have zero gradient
        between them. This is fixed with "penalty_matrix".

        Parameters
        ----------
        domains : dict
            The domains in which to compute the gradient matrix

        Returns
        -------
        :class:`pybamm.Matrix`
            The (sparse) Spectral Volume gradient matrix for the domain
        """
        submesh = self.mesh[domain]

        # Obtain the Chebyshev differentiation matrix.
        # Flip it, since it is defined for the Chebyshev
        # collocation points in descending order.
        chebdiff = np.flip(
            self.chebyshev_differentiation_matrices(self.order + 1, 1)[0]
        )

        # Create 1D matrix using submesh
        # submesh.npts is the number of CVs and n the number of SVs
        n = submesh.npts // self.order
        d = self.order
        # Compute the lengths of the Spectral Volumes.
        d_sv_edges = np.array(
            [
                np.sum(submesh.d_edges[d * i : d * i + d])
                for i in range(len(submesh.d_edges) // d)
            ]
        )
        # The 2 scales from [-1,1] (Chebyshev default) to [0,1].
        # e = 2 / submesh.d_sv_edges
        e = 2 / d_sv_edges
        # This factor scales the contribution of the reconstructed
        # gradient to the finite difference at the SV edges.
        # 0.0 is the value that makes it work with the "penalty_matrix".
        # 0.5 is the value that makes it work without it, but remember,
        # that effectively removes any implicit continuity conditions.
        f = 0.0
        # Here, the differentials are scaled to the SV.
        sub_matrix_raw = csr_matrix(kron(diags(e), chebdiff))
        if n == 1:
            sub_matrix = sub_matrix_raw
        else:
            sub_matrix = lil_matrix((n * d + 1, n * (d + 1)))
            sub_matrix[:d, : d + 1] = sub_matrix_raw[:d, : d + 1]
            sub_matrix[d, : d + 1] = f * sub_matrix_raw[d, : d + 1]
            # for loop of shame (optimisation potential via vectorisation)
            for i in range(1, n - 1):
                sub_matrix[i * d, i * (d + 1) : (i + 1) * (d + 1)] = (
                    f * sub_matrix_raw[i * (d + 1), i * (d + 1) : (i + 1) * (d + 1)]
                )
                sub_matrix[i * d + 1 : (i + 1) * d, i * (d + 1) : (i + 1) * (d + 1)] = (
                    sub_matrix_raw[
                        i * (d + 1) + 1 : (i + 1) * (d + 1) - 1,
                        i * (d + 1) : (i + 1) * (d + 1),
                    ]
                )
                sub_matrix[(i + 1) * d, i * (d + 1) : (i + 1) * (d + 1)] = (
                    f * sub_matrix_raw[i * (d + 1) + d, i * (d + 1) : (i + 1) * (d + 1)]
                )
            sub_matrix[-d - 1, -d - 1 :] = f * sub_matrix_raw[-d - 1, -d - 1 :]
            sub_matrix[-d:, -d - 1 :] = sub_matrix_raw[-d:, -d - 1 :]

        # number of repeats
        second_dim_repeats = self._get_auxiliary_domain_repeats(domains)

        # generate full matrix from the submatrix
        # Convert to csr_matrix so that we can take the index
        # (row-slicing), which is not supported by the default kron
        # format. Note that this makes column-slicing inefficient,
        # but this should not be an issue.
        matrix = csr_matrix(kron(eye(second_dim_repeats), sub_matrix))

        return pybamm.Matrix(matrix)

    def penalty_matrix(self, domains):
        """
        Penalty matrix for Spectral Volume in the appropriate domain.
        This works the same as the "gradient_matrix" of FiniteVolume
        does, just between SVs and not between CVs. Think of it as a
        continuity penalty.

        Parameters
        ----------
        domains : dict
            The domains in which to compute the gradient matrix

        Returns
        -------
        :class:`pybamm.Matrix`
            The (sparse) Spectral Volume penalty matrix for the domain
        """
        submesh = self.mesh[domains["primary"]]

        # Create 1D matrix using submesh
        n = submesh.npts
        d = self.order
        e = np.zeros(n - 1)
        e[d - 1 :: d] = 1 / submesh.d_nodes[d - 1 :: d]
        sub_matrix = vstack(
            [
                np.zeros((1, n)),
                diags([-e, e], [0, 1], shape=(n - 1, n)),
                np.zeros((1, n)),
            ]
        )

        # number of repeats
        second_dim_repeats = self._get_auxiliary_domain_repeats(domains)

        # generate full matrix from the submatrix
        # Convert to csr_matrix so that we can take the index
        # (row-slicing), which is not supported by the default kron
        # format. Note that this makes column-slicing inefficient, but
        # this should not be an issue.
        matrix = csr_matrix(kron(eye(second_dim_repeats), sub_matrix))

        return pybamm.Matrix(matrix)

    # def spectral_volume_internal_neumann_condition(
    #    self, left_symbol_disc, right_symbol_disc, left_mesh, right_mesh
    # ):
    #    """
    #    A method to find the internal neumann conditions between two
    #    symbols on adjacent subdomains. This method is never called,
    #    it's just here to show how a reconstructed gradient-based
    #    internal neumann_condition would look like.
    #    Parameters
    #    ----------
    #    left_symbol_disc : :class:`pybamm.Symbol`
    #        The discretised symbol on the left subdomain
    #    right_symbol_disc : :class:`pybamm.Symbol`
    #        The discretised symbol on the right subdomain
    #    left_mesh : list
    #        The mesh on the left subdomain
    #    right_mesh : list
    #        The mesh on the right subdomain
    #    """
    #
    #    second_dim_repeats = self._get_auxiliary_domain_repeats(
    #        left_symbol_disc.domains
    #    )
    #
    #    if second_dim_repeats != self._get_auxiliary_domain_repeats(
    #        right_symbol_disc.domains
    #    ):
    #        raise pybamm.DomainError(
    #            "Number of secondary points in subdomains do not match"
    #        )
    #
    #    # Use the Spectral Volume reconstruction and differentiation.
    #    left_reconstruction_matrix = self.cv_boundary_reconstruction_matrix(
    #        left_symbol_disc.domain,
    #        left_symbol_disc.auxiliary_domains
    #    )
    #    left_gradient_matrix = self.gradient_matrix(
    #        left_symbol_disc.domain,
    #        left_symbol_disc.auxiliary_domains
    #    ).entries[-1]
    #    left_matrix = left_gradient_matrix @ left_reconstruction_matrix
    #
    #    right_reconstruction_matrix = self.cv_boundary_reconstruction_matrix(
    #        right_symbol_disc.domain,
    #        right_symbol_disc.auxiliary_domains
    #    )
    #    right_gradient_matrix = self.gradient_matrix(
    #        right_symbol_disc.domain,
    #        right_symbol_disc.auxiliary_domains
    #    ).entries[0]
    #    right_matrix = right_gradient_matrix @ right_reconstruction_matrix
    #
    #    # Remove domains to avoid clash
    #    left_domain = left_symbol_disc.domain
    #    right_domain = right_symbol_disc.domain
    #    left_auxiliary_domains = left_symbol_disc.auxiliary_domains
    #    right_auxiliary_domains = right_symbol_disc.auxiliary_domains
    #    left_symbol_disc.clear_domains()
    #    right_symbol_disc.clear_domains()
    #
    #    # Spectral Volume derivative (i.e., the mean of the two
    #    # reconstructed gradients from each side)
    #    # Note that this is the version without "penalty_matrix".
    #    dy_dx = 0.5 * (right_matrix @ right_symbol_disc
    #                   + left_matrix @ left_symbol_disc)
    #
    #    # Change domains back
    #    left_symbol_disc.domain = left_domain
    #    right_symbol_disc.domain = right_domain
    #    left_symbol_disc.auxiliary_domains = left_auxiliary_domains
    #    right_symbol_disc.auxiliary_domains = right_auxiliary_domains
    #
    #    return dy_dx

    def replace_dirichlet_values(self, symbol, discretised_symbol, bcs):
        """
        Replace the reconstructed value at Dirichlet boundaries with the
        boundary condition.

        Parameters
        ----------
        symbol : :class:`pybamm.SpatialVariable`
            The variable to be discretised
        discretised_symbol : :class:`pybamm.Vector`
            Contains the discretised variable
        bcs : dict of tuples (:class:`pybamm.Scalar`, str)
            Dictionary (with keys "left" and "right") of boundary
            conditions. Each boundary condition consists of a value and
            a flag indicating its type (e.g. "Dirichlet")

        Returns
        -------
        :class:`pybamm.Symbol`
            `Matrix @ discretised_symbol + bcs_vector`. When evaluated,
            this gives the discretised_symbol, with its boundary values
            replaced by the Dirichlet boundary conditions.
        """
        # get relevant grid points
        domain = symbol.domain
        submesh = self.mesh[domain]

        # Prepare sizes
        n = (submesh.npts // self.order) * (self.order + 1)
        second_dim_repeats = self._get_auxiliary_domain_repeats(symbol.domains)

        lbc_value, lbc_type = bcs["left"]
        rbc_value, rbc_type = bcs["right"]

        # write boundary values into vectors of according shape
        if lbc_type == "Dirichlet":
            lbc_sub_matrix = coo_matrix(([1], ([0], [0])), shape=(n, 1))
            lbc_matrix = csr_matrix(kron(eye(second_dim_repeats), lbc_sub_matrix))
            if lbc_value.evaluates_to_number():
                left_bc = lbc_value * pybamm.Vector(np.ones(second_dim_repeats))
            else:
                left_bc = lbc_value
            lbc_vector = pybamm.Matrix(lbc_matrix) @ left_bc
        elif lbc_type == "Neumann":
            lbc_vector = pybamm.Vector(np.zeros(n * second_dim_repeats))
        else:
            raise ValueError(
                "boundary condition must be Dirichlet or Neumann, " f"not '{lbc_type}'"
            )

        if rbc_type == "Dirichlet":
            rbc_sub_matrix = coo_matrix(([1], ([n - 1], [0])), shape=(n, 1))
            rbc_matrix = csr_matrix(kron(eye(second_dim_repeats), rbc_sub_matrix))
            if rbc_value.evaluates_to_number():
                right_bc = rbc_value * pybamm.Vector(np.ones(second_dim_repeats))
            else:
                right_bc = rbc_value
            rbc_vector = pybamm.Matrix(rbc_matrix) @ right_bc
        elif rbc_type == "Neumann":
            rbc_vector = pybamm.Vector(np.zeros(n * second_dim_repeats))
        else:
            raise ValueError(
                "boundary condition must be Dirichlet or Neumann, " f"not '{rbc_type}'"
            )

        bcs_vector = lbc_vector + rbc_vector
        # Need to match the domain. E.g. in the case of the boundary
        # condition on the particle, the gradient has domain particle
        # but the bcs_vector has domain electrode, since it is a
        # function of the macroscopic variables
        bcs_vector.copy_domains(discretised_symbol)

        # Make matrix which makes "gaps" at the boundaries into which
        # the known Dirichlet values will be added. If the boundary
        # condition is not Dirichlet, it acts as identity.
        sub_matrix = diags(
            [int(lbc_type != "Dirichlet")]
            + [1 for i in range(n - 2)]
            + [int(rbc_type != "Dirichlet")]
        )

        # repeat matrix for secondary dimensions
        # Convert to csr_matrix so that we can take the index
        # (row-slicing), which is not supported by the default kron
        # format. Note that this makes column-slicing inefficient, but
        # this should not be an issue.
        matrix = csr_matrix(kron(eye(second_dim_repeats), sub_matrix))

        new_symbol = pybamm.Matrix(matrix) @ discretised_symbol + bcs_vector

        return new_symbol

    def replace_neumann_values(self, symbol, discretised_gradient, bcs):
        """
        Replace the known values of the gradient from Neumann boundary
        conditions into the discretised gradient.

        Parameters
        ----------
        symbol : :class:`pybamm.SpatialVariable`
            The variable to be discretised
        discretised_gradient : :class:`pybamm.Vector`
            Contains the discretised gradient of symbol
        bcs : dict of tuples (:class:`pybamm.Scalar`, str)
            Dictionary (with keys "left" and "right") of boundary
            conditions. Each boundary condition consists of a value and
            a flag indicating its type (e.g. "Dirichlet")

        Returns
        -------
        :class:`pybamm.Symbol`
            `Matrix @ discretised_gradient + bcs_vector`. When
            evaluated, this gives the discretised_gradient, with its
            boundary values replaced by the Neumann boundary conditions.
        """
        # get relevant grid points
        domain = symbol.domain
        submesh = self.mesh[domain]

        # Prepare sizes
        n = submesh.npts + 1
        second_dim_repeats = self._get_auxiliary_domain_repeats(symbol.domains)

        lbc_value, lbc_type = bcs["left"]
        rbc_value, rbc_type = bcs["right"]

        # Add any values from Neumann boundary conditions to the bcs vector
        if lbc_type == "Neumann":
            lbc_sub_matrix = coo_matrix(([1], ([0], [0])), shape=(n, 1))
            lbc_matrix = csr_matrix(kron(eye(second_dim_repeats), lbc_sub_matrix))
            if lbc_value.evaluates_to_number():
                left_bc = lbc_value * pybamm.Vector(np.ones(second_dim_repeats))
            else:
                left_bc = lbc_value
            lbc_vector = pybamm.Matrix(lbc_matrix) @ left_bc
        elif lbc_type == "Dirichlet":
            lbc_vector = pybamm.Vector(np.zeros(n * second_dim_repeats))
        else:
            raise ValueError(
                "boundary condition must be Dirichlet or Neumann, " f"not '{lbc_type}'"
            )

        if rbc_type == "Neumann":
            rbc_sub_matrix = coo_matrix(([1], ([n - 1], [0])), shape=(n, 1))
            rbc_matrix = csr_matrix(kron(eye(second_dim_repeats), rbc_sub_matrix))
            if rbc_value.evaluates_to_number():
                right_bc = rbc_value * pybamm.Vector(np.ones(second_dim_repeats))
            else:
                right_bc = rbc_value
            rbc_vector = pybamm.Matrix(rbc_matrix) @ right_bc
        elif rbc_type == "Dirichlet":
            rbc_vector = pybamm.Vector(np.zeros(n * second_dim_repeats))
        else:
            raise ValueError(
                "boundary condition must be Dirichlet or Neumann, " f"not '{rbc_type}'"
            )

        bcs_vector = lbc_vector + rbc_vector
        # Need to match the domain. E.g. in the case of the boundary
        # condition on the particle, the gradient has domain particle
        # but the bcs_vector has domain electrode, since it is a
        # function of the macroscopic variables
        bcs_vector.copy_domains(discretised_gradient)

        # Make matrix which makes "gaps" at the boundaries into which
        # the known Neumann values will be added. If the boundary
        # condition is not Neumann, it acts as identity.
        sub_matrix = diags(
            [int(lbc_type != "Neumann")]
            + [1 for i in range(n - 2)]
            + [int(rbc_type != "Neumann")]
        )

        # repeat matrix for secondary dimensions
        # Convert to csr_matrix so that we can take the index
        # (row-slicing), which is not supported by the default kron
        # format. Note that this makes column-slicing inefficient, but
        # this should not be an issue.
        matrix = csr_matrix(kron(eye(second_dim_repeats), sub_matrix))

        new_gradient = pybamm.Matrix(matrix) @ discretised_gradient + bcs_vector

        return new_gradient



#
# Processed Variable class
#
from typing import Optional
import casadi
import numpy as np
import pybamm
from scipy.integrate import cumulative_trapezoid
import xarray as xr
import bisect


class ProcessedVariable:
    """
    An object that can be evaluated at arbitrary (scalars or vectors) t and x, and
    returns the (interpolated) value of the base variable at that t and x.

    Parameters
    ----------
    base_variables : list of :class:`pybamm.Symbol`
        A list of base variables with a method `evaluate(t,y)`, each entry of which
        returns the value of that variable for that particular sub-solution.
        A Solution can be comprised of sub-solutions which are the solutions of
        different models.
        Note that this can be any kind of node in the expression tree, not
        just a :class:`pybamm.Variable`.
        When evaluated, returns an array of size (m,n)
    base_variables_casadi : list of :class:`casadi.Function`
        A list of casadi functions. When evaluated, returns the same thing as
        `base_Variable.evaluate` (but more efficiently).
    solution : :class:`pybamm.Solution`
        The solution object to be used to create the processed variables
    time_integral : :class:`pybamm.ProcessedVariableTimeIntegral`, optional
        Not none if the variable is to be time-integrated (default is None)
    """

    def __init__(
        self,
        base_variables,
        base_variables_casadi,
        solution,
        time_integral: Optional[pybamm.ProcessedVariableTimeIntegral] = None,
    ):
        self.base_variables = base_variables
        self.base_variables_casadi = base_variables_casadi

        self.all_ts = solution.all_ts
        self.all_ys = solution.all_ys
        self.all_yps = solution.all_yps
        self.all_inputs = solution.all_inputs
        self.all_inputs_casadi = solution.all_inputs_casadi

        self.mesh = base_variables[0].mesh
        self.domain = base_variables[0].domain
        self.domains = base_variables[0].domains
        self.time_integral = time_integral

        # Process spatial variables
        geometry = solution.all_models[0].geometry
        self.spatial_variables = {}
        for domain_level, domain_names in self.domains.items():
            variables = []
            for domain in domain_names:
                variables += list(geometry[domain].keys())
            self.spatial_variables[domain_level] = variables

        # Sensitivity starts off uninitialized, only set when called
        self._sensitivities = None
        self.all_solution_sensitivities = solution._all_sensitivities

        # Store time
        self.t_pts = solution.t

        # Evaluate base variable at initial time
        self.base_eval_shape = self.base_variables[0].shape
        self.base_eval_size = self.base_variables[0].size

        self._xr_array_raw = None
        self._entries_raw = None
        self._entries_for_interp_raw = None
        self._coords_raw = None

    def initialise(self):
        if self.entries_raw_initialized:
            return

        entries = self.observe_raw()

        t = self.t_pts
        entries_for_interp, coords = self._interp_setup(entries, t)

        self._entries_raw = entries
        self._entries_for_interp_raw = entries_for_interp
        self._coords_raw = coords

    def observe_and_interp(self, t, fill_value):
        """
        Interpolate the variable at the given time points and y values.
        t must be a sorted array of time points.
        """

        entries = self._observe_hermite_cpp(t)
        processed_entries = self._observe_postfix(entries, t)

        tf = self.t_pts[-1]
        if t[-1] > tf and fill_value != "extrapolate":
            # fill the rest
            idx = np.searchsorted(t, tf, side="right")
            processed_entries[..., idx:] = fill_value

        return processed_entries

    def observe_raw(self):
        """
        Evaluate the base variable at the given time points and y values.
        """
        t = self.t_pts

        # For small number of points, use Python
        if pybamm.has_idaklu():
            entries = self._observe_raw_cpp()
        else:
            # Fallback method for when IDAKLU is not available. To be removed
            # when the C++ code is migrated to a new repo
            entries = self._observe_raw_python()  # pragma: no cover

        return self._observe_postfix(entries, t)

    def _setup_cpp_inputs(self, t, full_range):
        pybamm.logger.debug("Setting up C++ interpolation inputs")

        ts = self.all_ts
        ys = self.all_ys
        yps = self.all_yps
        inputs = self.all_inputs_casadi

        # Remove all empty ts
        idxs = np.where([ti.size > 0 for ti in ts])[0]

        # Find the indices of the time points to observe
        if not full_range:
            ts_nonempty = [ts[idx] for idx in idxs]
            idxs_subset = _find_ts_indices(ts_nonempty, t)
            idxs = idxs[idxs_subset]

        # Extract the time points and inputs
        ts = [ts[idx] for idx in idxs]
        ys = [ys[idx] for idx in idxs]
        if self.hermite_interpolation:
            yps = [yps[idx] for idx in idxs]
        inputs = [self.all_inputs_casadi[idx] for idx in idxs]

        is_f_contiguous = _is_f_contiguous(ys)

        ts = pybamm.solvers.idaklu_solver.idaklu.VectorRealtypeNdArray(ts)
        ys = pybamm.solvers.idaklu_solver.idaklu.VectorRealtypeNdArray(ys)
        if self.hermite_interpolation:
            yps = pybamm.solvers.idaklu_solver.idaklu.VectorRealtypeNdArray(yps)
        else:
            yps = None
        inputs = pybamm.solvers.idaklu_solver.idaklu.VectorRealtypeNdArray(inputs)

        # Generate the serialized C++ functions only once
        funcs_unique = {}
        funcs = [None] * len(idxs)
        for i in range(len(idxs)):
            vars = self.base_variables_casadi[idxs[i]]
            if vars not in funcs_unique:
                funcs_unique[vars] = vars.serialize()
            funcs[i] = funcs_unique[vars]

        return ts, ys, yps, funcs, inputs, is_f_contiguous

    def _observe_hermite_cpp(self, t):
        pybamm.logger.debug("Observing and Hermite interpolating the variable in C++")

        ts, ys, yps, funcs, inputs, _ = self._setup_cpp_inputs(t, full_range=False)
        shapes = self._shape(t)
        return pybamm.solvers.idaklu_solver.idaklu.observe_hermite_interp(
            t, ts, ys, yps, inputs, funcs, shapes
        )

    def _observe_raw_cpp(self):
        pybamm.logger.debug("Observing the variable raw data in C++")
        t = self.t_pts
        ts, ys, _, funcs, inputs, is_f_contiguous = self._setup_cpp_inputs(
            t, full_range=True
        )
        shapes = self._shape(self.t_pts)

        return pybamm.solvers.idaklu_solver.idaklu.observe(
            ts, ys, inputs, funcs, is_f_contiguous, shapes
        )

    def _observe_raw_python(self):
        raise NotImplementedError  # pragma: no cover

    def _observe_postfix(self, entries, t):
        return entries

    def _interp_setup(self, entries, t):
        raise NotImplementedError  # pragma: no cover

    def _shape(self, t):
        raise NotImplementedError  # pragma: no cover

    def _process_spatial_variable_names(self, spatial_variable):
        if len(spatial_variable) == 0:
            return None

        # Extract names
        raw_names = []
        for var in spatial_variable:
            # Ignore tabs in domain names
            if var == "tabs":
                continue
            if isinstance(var, str):
                raw_names.append(var)
            else:
                raw_names.append(var.name)

        # Rename battery variables to match PyBaMM convention
        if all([var.startswith("r") for var in raw_names]):
            return "r"
        elif all([var.startswith("x") for var in raw_names]):
            return "x"
        elif all([var.startswith("R") for var in raw_names]):
            return "R"
        elif len(raw_names) == 1:
            return raw_names[0]
        else:
            raise NotImplementedError(
                f"Spatial variable name not recognized for {spatial_variable}"
            )

    def __call__(
        self,
        t=None,
        x=None,
        r=None,
        y=None,
        z=None,
        R=None,
        fill_value=np.nan,
    ):
        # Check to see if we are interpolating exactly onto the raw solution time points
        t_observe, observe_raw = self._check_observe_raw(t)

        # Check if the time points are sorted and unique
        is_sorted = observe_raw or _is_sorted(t_observe)

        # Sort them if not
        if not is_sorted:
            idxs_sort = np.argsort(t_observe)
            t_observe = t_observe[idxs_sort]

        hermite_time_interp = (
            pybamm.has_idaklu() and self.hermite_interpolation and not observe_raw
        )

        if hermite_time_interp:
            entries = self.observe_and_interp(t_observe, fill_value)

        spatial_interp = any(a is not None for a in [x, r, y, z, R])

        xr_interp = spatial_interp or not hermite_time_interp

        if xr_interp:
            if hermite_time_interp:
                # Already interpolated in time
                t = None
                entries_for_interp, coords = self._interp_setup(entries, t_observe)
            else:
                self.initialise()
                entries_for_interp, coords = (
                    self._entries_for_interp_raw,
                    self._coords_raw,
                )

            if self.time_integral is None:
                processed_entries = self._xr_interpolate(
                    entries_for_interp,
                    coords,
                    observe_raw,
                    t,
                    x,
                    r,
                    y,
                    z,
                    R,
                    fill_value,
                )
            else:
                processed_entries = entries_for_interp
        else:
            processed_entries = entries

        if not is_sorted:
            idxs_unsort = np.zeros_like(idxs_sort)
            idxs_unsort[idxs_sort] = np.arange(len(t_observe))

            processed_entries = processed_entries[..., idxs_unsort]

        # Remove a singleton time dimension if we interpolate in time with hermite
        if hermite_time_interp and t_observe.size == 1:
            processed_entries = np.squeeze(processed_entries, axis=-1)

        return processed_entries

    def _xr_interpolate(
        self,
        entries_for_interp,
        coords,
        observe_raw,
        t=None,
        x=None,
        r=None,
        y=None,
        z=None,
        R=None,
        fill_value=None,
    ):
        """
        Evaluate the variable at arbitrary *dimensional* t (and x, r, y, z and/or R),
        using interpolation
        """
        if observe_raw:
            if not self.xr_array_raw_initialized:
                self._xr_array_raw = xr.DataArray(entries_for_interp, coords=coords)
            xr_data_array = self._xr_array_raw
        else:
            xr_data_array = xr.DataArray(entries_for_interp, coords=coords)

        kwargs = {"t": t, "x": x, "r": r, "y": y, "z": z, "R": R}

        # Remove any None arguments
        kwargs = {key: value for key, value in kwargs.items() if value is not None}

        # Use xarray interpolation, return numpy array
        out = xr_data_array.interp(**kwargs, kwargs={"fill_value": fill_value}).values

        return out

    def _check_observe_raw(self, t):
        """
        Checks if the raw data should be observed exactly at the solution time points

        Args:
            t (np.ndarray, list, None): time points to observe

        Returns:
            t_observe (np.ndarray): time points to observe
            observe_raw (bool): True if observing the raw data
        """
        # if this is a time integral variable, t must be None and we observe either the
        # data times (for a discrete sum) or the solution times (for a continuous sum)
        if self.time_integral is not None:
            if self.time_integral.method == "discrete":
                # discrete sum should be observed at the discrete times
                t = self.time_integral.discrete_times
            else:
                # assume we can do a sufficiently accurate trapezoidal integration at t_pts
                t = self.t_pts

        observe_raw = (t is None) or (
            np.asarray(t).size == len(self.t_pts) and np.all(t == self.t_pts)
        )

        if observe_raw:
            t_observe = self.t_pts
        elif not isinstance(t, np.ndarray):
            if not isinstance(t, list):
                t = [t]
            t_observe = np.array(t)
        else:
            t_observe = t

        if t_observe[0] < self.t_pts[0]:
            raise ValueError(
                "The interpolation points must be greater than or equal to the initial solution time"
            )

        return t_observe, observe_raw

    @property
    def entries(self):
        """
        Returns the raw data entries of the processed variable. If the processed
        variable has not been initialized (i.e. the entries have not been
        calculated), then the processed variable is initialized first.
        """
        self.initialise()
        return self._entries_raw

    @property
    def data(self):
        """Same as entries, but different name"""
        return self.entries

    @property
    def entries_raw_initialized(self):
        return self._entries_raw is not None

    @property
    def xr_array_raw_initialized(self):
        return self._xr_array_raw is not None

    @property
    def sensitivities(self):
        """
        Returns a dictionary of sensitivities for each input parameter.
        The keys are the input parameters, and the value is a matrix of size
        (n_x * n_t, n_p), where n_x is the number of states, n_t is the number of time
        points, and n_p is the size of the input parameter
        """
        # No sensitivities if there are no inputs
        if len(self.all_inputs[0]) == 0:
            return {}
        # Otherwise initialise and return sensitivities
        if self._sensitivities is None:
            if self.all_solution_sensitivities:
                self.initialise_sensitivity_explicit_forward()
            else:
                raise ValueError(
                    "Cannot compute sensitivities. The 'calculate_sensitivities' "
                    "argument of the solver.solve should be changed from 'None' to "
                    "allow sensitivities calculations. Check solver documentation for "
                    "details."
                )
        return self._sensitivities

    def initialise_sensitivity_explicit_forward(self):
        "Set up the sensitivity dictionary"

        all_S_var = []
        for ts, ys, inputs_stacked, inputs, base_variable, dy_dp in zip(
            self.all_ts,
            self.all_ys,
            self.all_inputs_casadi,
            self.all_inputs,
            self.base_variables,
            self.all_solution_sensitivities["all"],
        ):
            # Set up symbolic variables
            t_casadi = casadi.MX.sym("t")
            y_casadi = casadi.MX.sym("y", ys.shape[0])
            p_casadi = {
                name: casadi.MX.sym(name, value.shape[0])
                for name, value in inputs.items()
            }

            p_casadi_stacked = casadi.vertcat(*[p for p in p_casadi.values()])

            # Convert variable to casadi format for differentiating
            var_casadi = base_variable.to_casadi(t_casadi, y_casadi, inputs=p_casadi)
            dvar_dy = casadi.jacobian(var_casadi, y_casadi)
            dvar_dp = casadi.jacobian(var_casadi, p_casadi_stacked)

            # Convert to functions and evaluate index-by-index
            dvar_dy_func = casadi.Function(
                "dvar_dy", [t_casadi, y_casadi, p_casadi_stacked], [dvar_dy]
            )
            dvar_dp_func = casadi.Function(
                "dvar_dp", [t_casadi, y_casadi, p_casadi_stacked], [dvar_dp]
            )
            for idx, t in enumerate(ts):
                u = ys[:, idx]
                next_dvar_dy_eval = dvar_dy_func(t, u, inputs_stacked)
                next_dvar_dp_eval = dvar_dp_func(t, u, inputs_stacked)
                if idx == 0:
                    dvar_dy_eval = next_dvar_dy_eval
                    dvar_dp_eval = next_dvar_dp_eval
                else:
                    dvar_dy_eval = casadi.diagcat(dvar_dy_eval, next_dvar_dy_eval)
                    dvar_dp_eval = casadi.vertcat(dvar_dp_eval, next_dvar_dp_eval)

            # Compute sensitivity
            S_var = dvar_dy_eval @ dy_dp + dvar_dp_eval
            all_S_var.append(S_var)

        S_var = casadi.vertcat(*all_S_var)
        sensitivities = {"all": S_var}

        # Add the individual sensitivity
        start = 0
        for name, inp in self.all_inputs[0].items():
            end = start + inp.shape[0]
            sensitivities[name] = S_var[:, start:end]
            start = end

        # Save attribute
        self._sensitivities = sensitivities

    @property
    def hermite_interpolation(self):
        return self.all_yps is not None


class ProcessedVariable0D(ProcessedVariable):
    def __init__(
        self,
        base_variables,
        base_variables_casadi,
        solution,
        time_integral: Optional[pybamm.ProcessedVariableTimeIntegral] = None,
    ):
        self.dimensions = 0
        super().__init__(
            base_variables,
            base_variables_casadi,
            solution,
            time_integral=time_integral,
        )

    def _observe_raw_python(self):
        pybamm.logger.debug("Observing the variable raw data in Python")
        # initialise empty array of the correct size
        entries = np.empty(self._shape(self.t_pts))
        idx = 0
        # Evaluate the base_variable index-by-index
        for ts, ys, inputs, base_var_casadi in zip(
            self.all_ts, self.all_ys, self.all_inputs_casadi, self.base_variables_casadi
        ):
            for inner_idx, t in enumerate(ts):
                t = ts[inner_idx]
                y = ys[:, inner_idx]
                entries[idx] = float(base_var_casadi(t, y, inputs))

                idx += 1
        return entries

    def _observe_postfix(self, entries, t):
        if self.time_integral is None:
            return entries
        if self.time_integral.method == "discrete":
            return np.sum(entries, axis=0, initial=self.time_integral.initial_condition)
        elif self.time_integral.method == "continuous":
            return cumulative_trapezoid(
                entries, self.t_pts, initial=float(self.time_integral.initial_condition)
            )
        else:
            raise ValueError(
                "time_integral method must be 'discrete' or 'continuous'"
            )  # pragma: no cover

    def _interp_setup(self, entries, t):
        # save attributes for interpolation
        entries_for_interp = entries
        coords_for_interp = {"t": t}

        return entries_for_interp, coords_for_interp

    def _shape(self, t):
        return [len(t)]


class ProcessedVariable1D(ProcessedVariable):
    """
    An object that can be evaluated at arbitrary (scalars or vectors) t and x, and
    returns the (interpolated) value of the base variable at that t and x.

    Parameters
    ----------
    base_variables : list of :class:`pybamm.Symbol`
        A list of base variables with a method `evaluate(t,y)`, each entry of which
        returns the value of that variable for that particular sub-solution.
        A Solution can be comprised of sub-solutions which are the solutions of
        different models.
        Note that this can be any kind of node in the expression tree, not
        just a :class:`pybamm.Variable`.
        When evaluated, returns an array of size (m,n)
    base_variables_casadi : list of :class:`casadi.Function`
        A list of casadi functions. When evaluated, returns the same thing as
        `base_Variable.evaluate` (but more efficiently).
    solution : :class:`pybamm.Solution`
        The solution object to be used to create the processed variables
    """

    def __init__(
        self,
        base_variables,
        base_variables_casadi,
        solution,
        time_integral: Optional[pybamm.ProcessedVariableTimeIntegral] = None,
    ):
        self.dimensions = 1
        super().__init__(
            base_variables,
            base_variables_casadi,
            solution,
            time_integral=time_integral,
        )

    def _observe_raw_python(self):
        pybamm.logger.debug("Observing the variable raw data in Python")
        entries = np.empty(self._shape(self.t_pts))

        # Evaluate the base_variable index-by-index
        idx = 0
        for ts, ys, inputs, base_var_casadi in zip(
            self.all_ts, self.all_ys, self.all_inputs_casadi, self.base_variables_casadi
        ):
            for inner_idx, t in enumerate(ts):
                t = ts[inner_idx]
                y = ys[:, inner_idx]
                entries[:, idx] = base_var_casadi(t, y, inputs).full()[:, 0]
                idx += 1
        return entries

    def _interp_setup(self, entries, t):
        # Get node and edge values
        nodes = self.mesh.nodes
        edges = self.mesh.edges
        if entries.shape[0] == len(nodes):
            space = nodes
        elif entries.shape[0] == len(edges):
            space = edges

        # add points outside domain for extrapolation to boundaries
        extrap_space_left = np.array([2 * space[0] - space[1]])
        extrap_space_right = np.array([2 * space[-1] - space[-2]])
        space = np.concatenate([extrap_space_left, space, extrap_space_right])
        extrap_entries_left = 2 * entries[0] - entries[1]
        extrap_entries_right = 2 * entries[-1] - entries[-2]
        entries_for_interp = np.vstack(
            [extrap_entries_left, entries, extrap_entries_right]
        )

        # assign attributes for reference (either x_sol or r_sol)
        self.spatial_variable_names = {
            k: self._process_spatial_variable_names(v)
            for k, v in self.spatial_variables.items()
        }
        self.first_dimension = self.spatial_variable_names["primary"]

        # assign attributes for reference
        pts_for_interp = space
        self.internal_boundaries = self.mesh.internal_boundaries

        # Set first_dim_pts to edges for nicer plotting
        self.first_dim_pts = edges

        # save attributes for interpolation
        coords_for_interp = {self.first_dimension: pts_for_interp, "t": t}

        return entries_for_interp, coords_for_interp

    def _shape(self, t):
        t_size = len(t)
        space_size = self.base_eval_shape[0]
        return [space_size, t_size]


class ProcessedVariable2D(ProcessedVariable):
    """
    An object that can be evaluated at arbitrary (scalars or vectors) t and x, and
    returns the (interpolated) value of the base variable at that t and x.

    Parameters
    ----------
    base_variables : list of :class:`pybamm.Symbol`
        A list of base variables with a method `evaluate(t,y)`, each entry of which
        returns the value of that variable for that particular sub-solution.
        A Solution can be comprised of sub-solutions which are the solutions of
        different models.
        Note that this can be any kind of node in the expression tree, not
        just a :class:`pybamm.Variable`.
        When evaluated, returns an array of size (m,n)
    base_variables_casadi : list of :class:`casadi.Function`
        A list of casadi functions. When evaluated, returns the same thing as
        `base_Variable.evaluate` (but more efficiently).
    solution : :class:`pybamm.Solution`
        The solution object to be used to create the processed variables
    """

    def __init__(
        self,
        base_variables,
        base_variables_casadi,
        solution,
        time_integral: Optional[pybamm.ProcessedVariableTimeIntegral] = None,
    ):
        self.dimensions = 2
        super().__init__(
            base_variables,
            base_variables_casadi,
            solution,
            time_integral=time_integral,
        )
        first_dim_nodes = self.mesh.nodes
        first_dim_edges = self.mesh.edges
        second_dim_nodes = self.base_variables[0].secondary_mesh.nodes
        if self.base_eval_size // len(second_dim_nodes) == len(first_dim_nodes):
            first_dim_pts = first_dim_nodes
        elif self.base_eval_size // len(second_dim_nodes) == len(first_dim_edges):
            first_dim_pts = first_dim_edges

        second_dim_pts = second_dim_nodes
        self.first_dim_size = len(first_dim_pts)
        self.second_dim_size = len(second_dim_pts)

    def _observe_raw_python(self):
        """
        Initialise a 2D object that depends on x and r, x and z, x and R, or R and r.
        """
        pybamm.logger.debug("Observing the variable raw data in Python")
        first_dim_size, second_dim_size, t_size = self._shape(self.t_pts)
        entries = np.empty((first_dim_size, second_dim_size, t_size))

        # Evaluate the base_variable index-by-index
        idx = 0
        for ts, ys, inputs, base_var_casadi in zip(
            self.all_ts, self.all_ys, self.all_inputs_casadi, self.base_variables_casadi
        ):
            for inner_idx, t in enumerate(ts):
                t = ts[inner_idx]
                y = ys[:, inner_idx]
                entries[:, :, idx] = np.reshape(
                    base_var_casadi(t, y, inputs).full(),
                    [first_dim_size, second_dim_size],
                    order="F",
                )
                idx += 1
        return entries

    def _interp_setup(self, entries, t):
        """
        Initialise a 2D object that depends on x and r, x and z, x and R, or R and r.
        """
        first_dim_nodes = self.mesh.nodes
        first_dim_edges = self.mesh.edges
        second_dim_nodes = self.base_variables[0].secondary_mesh.nodes
        second_dim_edges = self.base_variables[0].secondary_mesh.edges
        if self.base_eval_size // len(second_dim_nodes) == len(first_dim_nodes):
            first_dim_pts = first_dim_nodes
        elif self.base_eval_size // len(second_dim_nodes) == len(first_dim_edges):
            first_dim_pts = first_dim_edges

        second_dim_pts = second_dim_nodes

        # add points outside first dimension domain for extrapolation to
        # boundaries
        extrap_space_first_dim_left = np.array(
            [2 * first_dim_pts[0] - first_dim_pts[1]]
        )
        extrap_space_first_dim_right = np.array(
            [2 * first_dim_pts[-1] - first_dim_pts[-2]]
        )
        first_dim_pts = np.concatenate(
            [extrap_space_first_dim_left, first_dim_pts, extrap_space_first_dim_right]
        )
        extrap_entries_left = np.expand_dims(2 * entries[0] - entries[1], axis=0)
        extrap_entries_right = np.expand_dims(2 * entries[-1] - entries[-2], axis=0)
        entries_for_interp = np.concatenate(
            [extrap_entries_left, entries, extrap_entries_right], axis=0
        )

        # add points outside second dimension domain for extrapolation to
        # boundaries
        extrap_space_second_dim_left = np.array(
            [2 * second_dim_pts[0] - second_dim_pts[1]]
        )
        extrap_space_second_dim_right = np.array(
            [2 * second_dim_pts[-1] - second_dim_pts[-2]]
        )
        second_dim_pts = np.concatenate(
            [
                extrap_space_second_dim_left,
                second_dim_pts,
                extrap_space_second_dim_right,
            ]
        )
        extrap_entries_second_dim_left = np.expand_dims(
            2 * entries_for_interp[:, 0, :] - entries_for_interp[:, 1, :], axis=1
        )
        extrap_entries_second_dim_right = np.expand_dims(
            2 * entries_for_interp[:, -1, :] - entries_for_interp[:, -2, :], axis=1
        )
        entries_for_interp = np.concatenate(
            [
                extrap_entries_second_dim_left,
                entries_for_interp,
                extrap_entries_second_dim_right,
            ],
            axis=1,
        )

        self.spatial_variable_names = {
            k: self._process_spatial_variable_names(v)
            for k, v in self.spatial_variables.items()
        }

        self.first_dimension = self.spatial_variable_names["primary"]
        self.second_dimension = self.spatial_variable_names["secondary"]

        # assign attributes for reference
        first_dim_pts_for_interp = first_dim_pts
        second_dim_pts_for_interp = second_dim_pts

        # Set pts to edges for nicer plotting
        self.first_dim_pts = first_dim_edges
        self.second_dim_pts = second_dim_edges

        # save attributes for interpolation
        coords_for_interp = {
            self.first_dimension: first_dim_pts_for_interp,
            self.second_dimension: second_dim_pts_for_interp,
            "t": t,
        }

        return entries_for_interp, coords_for_interp

    def _shape(self, t):
        first_dim_size = self.first_dim_size
        second_dim_size = self.second_dim_size
        t_size = len(t)
        return [first_dim_size, second_dim_size, t_size]


class ProcessedVariable2DSciKitFEM(ProcessedVariable2D):
    """
    An object that can be evaluated at arbitrary (scalars or vectors) t and x, and
    returns the (interpolated) value of the base variable at that t and x.

    Parameters
    ----------
    base_variables : list of :class:`pybamm.Symbol`
        A list of base variables with a method `evaluate(t,y)`, each entry of which
        returns the value of that variable for that particular sub-solution.
        A Solution can be comprised of sub-solutions which are the solutions of
        different models.
        Note that this can be any kind of node in the expression tree, not
        just a :class:`pybamm.Variable`.
        When evaluated, returns an array of size (m,n)
    base_variables_casadi : list of :class:`casadi.Function`
        A list of casadi functions. When evaluated, returns the same thing as
        `base_Variable.evaluate` (but more efficiently).
    solution : :class:`pybamm.Solution`
        The solution object to be used to create the processed variables
    """

    def __init__(
        self,
        base_variables,
        base_variables_casadi,
        solution,
        time_integral: Optional[pybamm.ProcessedVariableTimeIntegral] = None,
    ):
        self.dimensions = 2
        super(ProcessedVariable2D, self).__init__(
            base_variables,
            base_variables_casadi,
            solution,
            time_integral=time_integral,
        )
        y_sol = self.mesh.edges["y"]
        z_sol = self.mesh.edges["z"]

        self.first_dim_size = len(y_sol)
        self.second_dim_size = len(z_sol)

    def _interp_setup(self, entries, t):
        y_sol = self.mesh.edges["y"]
        z_sol = self.mesh.edges["z"]

        # assign attributes for reference
        self.y_sol = y_sol
        self.z_sol = z_sol
        self.first_dimension = "y"
        self.second_dimension = "z"
        self.first_dim_pts = y_sol
        self.second_dim_pts = z_sol

        # save attributes for interpolation
        coords_for_interp = {"y": y_sol, "z": z_sol, "t": t}

        return entries, coords_for_interp


def process_variable(base_variables, *args, **kwargs):
    mesh = base_variables[0].mesh
    domain = base_variables[0].domain

    # Evaluate base variable at initial time
    base_eval_shape = base_variables[0].shape
    base_eval_size = base_variables[0].size

    # handle 2D (in space) finite element variables differently
    if (
        mesh
        and "current collector" in domain
        and isinstance(mesh, pybamm.ScikitSubMesh2D)
    ):
        return ProcessedVariable2DSciKitFEM(base_variables, *args, **kwargs)

    # check variable shape
    if len(base_eval_shape) == 0 or base_eval_shape[0] == 1:
        return ProcessedVariable0D(base_variables, *args, **kwargs)

    n = mesh.npts
    base_shape = base_eval_shape[0]
    # Try some shapes that could make the variable a 1D variable
    if base_shape in [n, n + 1]:
        return ProcessedVariable1D(base_variables, *args, **kwargs)

    # Try some shapes that could make the variable a 2D variable
    first_dim_nodes = mesh.nodes
    first_dim_edges = mesh.edges
    second_dim_pts = base_variables[0].secondary_mesh.nodes
    if base_eval_size // len(second_dim_pts) in [
        len(first_dim_nodes),
        len(first_dim_edges),
    ]:
        return ProcessedVariable2D(base_variables, *args, **kwargs)

    # Raise error for 3D variable
    raise NotImplementedError(
        f"Shape not recognized for {base_variables[0]}"
        + "(note processing of 3D variables is not yet implemented)"
    )


def _is_f_contiguous(all_ys):
    """
    Check if all the ys are f-contiguous in memory

    Args:
        all_ys (list of np.ndarray): list of all ys

    Returns:
        bool: True if all ys are f-contiguous
    """

    return all(isinstance(y, np.ndarray) and y.data.f_contiguous for y in all_ys)


def _is_sorted(t):
    """
    Check if an array is sorted

    Args:
        t (np.ndarray): array to check

    Returns:
        bool: True if array is sorted
    """
    return np.all(t[:-1] <= t[1:])


def _find_ts_indices(ts, t):
    """
    Parameters:
    - ts: A list of numpy arrays (each sorted) whose values are successively increasing.
    - t: A sorted list or array of values to find within ts.

    Returns:
    - indices: A list of indices from `ts` such that at least one value of `t` falls within ts[idx].
    """

    indices = []

    # Get the minimum and maximum values of the target values `t`
    t_min, t_max = t[0], t[-1]

    # Step 1: Use binary search to find the range of `ts` arrays where t_min and t_max could lie
    low_idx = bisect.bisect_left([ts_arr[-1] for ts_arr in ts], t_min)
    high_idx = bisect.bisect_right([ts_arr[0] for ts_arr in ts], t_max)

    # Step 2: Iterate over the identified range
    for idx in range(low_idx, high_idx):
        ts_min, ts_max = ts[idx][0], ts[idx][-1]

        # Binary search within `t` to check if any value falls within [ts_min, ts_max]
        i = bisect.bisect_left(t, ts_min)
        if i < len(t) and t[i] <= ts_max:
            # At least one value of t is within ts[idx]
            indices.append(idx)

    # extrapolating
    if (t[-1] > ts[-1][-1]) and (len(indices) == 0 or indices[-1] != len(ts) - 1):
        indices.append(len(ts) - 1)

    return indices



import casadi
import pybamm
import numpy as np
import warnings
from scipy.interpolate import interp1d
from .lrudict import LRUDict


class CasadiSolver(pybamm.BaseSolver):
    """Solve a discretised model, using CasADi.

    Parameters
    ----------
    mode : str
        How to solve the model (default is "safe"):

            - "fast": perform direct integration, without accounting for events. \
            Recommended when simulating a drive cycle or other simulation where \
            no events should be triggered.
            - "fast with events": perform direct integration of the whole timespan, \
            then go back and check where events were crossed. Experimental only.
            - "safe": perform step-and-check integration in global steps of size \
            dt_max, checking whether events have been triggered. Recommended for \
            simulations of a full charge or discharge.
            - "safe without grid": perform step-and-check integration step-by-step. \
            Takes more steps than "safe" mode, but doesn't require creating the grid \
            each time, so may be faster. Experimental only.
    rtol : float, optional
        The relative tolerance for the solver (default is 1e-6).
    atol : float, optional
        The absolute tolerance for the solver (default is 1e-6).
    root_method : str or pybamm algebraic solver class, optional
        The method to use to find initial conditions (for DAE solvers).
        If a solver class, must be an algebraic solver class.
        If "casadi",
        the solver uses casadi's Newton rootfinding algorithm to find initial
        conditions. Otherwise, the solver uses 'scipy.optimize.root' with method
        specified by 'root_method' (e.g. "lm", "hybr", ...)
    root_tol : float, optional
        The tolerance for root-finding. Default is 1e-6.
    max_step_decrease_count : float, optional
        The maximum number of times step size can be decreased before an error is
        raised. Default is 5.
    dt_max : float, optional
        The maximum global step size (in seconds) used in "safe" mode. If None
        the default value is 600 seconds.
    extrap_tol : float, optional
        The tolerance to assert whether extrapolation occurs or not. Default is 0.
    extra_options_setup : dict, optional
        Any options to pass to the CasADi integrator when creating the integrator.
        Please consult `CasADi documentation <https://web.casadi.org/python-api/#integrator>`_ for
        details. Some useful options:

        - "max_num_steps": Maximum number of integrator steps
        - "print_stats": Print out statistics after integration

    extra_options_call : dict, optional
        Any options to pass to the CasADi integrator when calling the integrator.
        Please consult `CasADi documentation <https://web.casadi.org/python-api/#integrator>`_ for
        details.
    return_solution_if_failed_early : bool, optional
        Whether to return a Solution object if the solver fails to reach the end of
        the simulation, but managed to take some successful steps. Default is False.
    perturb_algebraic_initial_conditions : bool, optional
        Whether to perturb algebraic initial conditions to avoid a singularity. This
        can sometimes slow down the solver, but is kept True as default for "safe" mode
        as it seems to be more robust (False by default for other modes).
    integrators_maxcount : int, optional
        The maximum number of integrators that the solver will retain before
        ejecting past integrators using an LRU methodology. A value of 0 or
        None leaves the number of integrators unbound. Default is 100.
    """

    def __init__(
        self,
        mode="safe",
        rtol=1e-6,
        atol=1e-6,
        root_method="casadi",
        root_tol=1e-6,
        max_step_decrease_count=5,
        dt_max=None,
        extrap_tol=None,
        extra_options_setup=None,
        extra_options_call=None,
        return_solution_if_failed_early=False,
        perturb_algebraic_initial_conditions=None,
        integrators_maxcount=100,
    ):
        super().__init__(
            "problem dependent",
            rtol,
            atol,
            root_method,
            root_tol,
            extrap_tol,
        )
        if mode in ["safe", "fast", "fast with events", "safe without grid"]:
            self.mode = mode
        else:
            raise ValueError(
                f"invalid mode '{mode}'. Must be 'safe', for solving with events, "
                "'fast', for solving quickly without events, or 'safe without grid' or "
                "'fast with events' (both experimental)"
            )
        self.max_step_decrease_count = max_step_decrease_count
        self.dt_max = dt_max or 600

        self.extra_options_setup = extra_options_setup or {}
        self.extra_options_call = extra_options_call or {}
        self.return_solution_if_failed_early = return_solution_if_failed_early

        self._on_extrapolation = "error"

        # Decide whether to perturb algebraic initial conditions, True by default for
        # "safe" mode, False by default for other modes
        if perturb_algebraic_initial_conditions is None:
            if mode == "safe":
                self.perturb_algebraic_initial_conditions = True
            else:
                self.perturb_algebraic_initial_conditions = False
        else:
            self.perturb_algebraic_initial_conditions = (
                perturb_algebraic_initial_conditions
            )
        self.name = f"CasADi solver with '{mode}' mode"

        # Initialize
        self.integrators_maxcount = integrators_maxcount
        self.integrators = LRUDict(maxsize=self.integrators_maxcount)
        self.integrator_specs = LRUDict(maxsize=self.integrators_maxcount)
        self.y_sols = {}

        pybamm.citations.register("Andersson2019")

    def _integrate(self, model, t_eval, inputs_dict=None, t_interp=None):
        """
        Solve a DAE model defined by residuals with initial conditions y0.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        t_eval : numeric type
            The times at which to compute the solution
        inputs_dict : dict, optional
            Any input parameters to pass to the model when solving
        """

        # Record whether there are any symbolic inputs
        inputs_dict = inputs_dict or {}

        # convert inputs to casadi format
        inputs = casadi.vertcat(*[x for x in inputs_dict.values()])

        if self.mode in ["fast", "fast with events"] or not model.events:
            if not model.events:
                pybamm.logger.info("No events found, running fast mode")
            if self.mode == "fast with events":
                # Create the integrator with an event switch that will set the rhs to
                # zero when voltage limits are crossed
                use_event_switch = True
            else:
                use_event_switch = False
            # Create an integrator with the grid (we just need to do this once)
            self.create_integrator(
                model, inputs, t_eval, use_event_switch=use_event_switch
            )
            solution = self._run_integrator(
                model, model.y0, inputs_dict, inputs, t_eval
            )
            # Check if the sign of an event changes, if so find an accurate
            # termination point and exit
            solution = self._solve_for_event(solution)
            solution.check_ys_are_not_too_large()
            return solution
        elif self.mode in ["safe", "safe without grid"]:
            y0 = model.y0
            # Step-and-check
            t = t_eval[0]
            t_f = t_eval[-1]

            pybamm.logger.debug(f"Start solving {model.name} with {self.name}")

            if self.mode == "safe without grid":
                # in "safe without grid" mode,
                # create integrator once, without grid,
                # to avoid having to create several times
                self.create_integrator(model, inputs)
                # Initialize solution
                solution = pybamm.Solution(
                    np.array([t]),
                    y0,
                    model,
                    inputs_dict,
                    all_sensitivities=False,
                )
                solution.solve_time = 0
                solution.integration_time = 0
                use_grid = False
            else:
                solution = None
                use_grid = True

            # Try to integrate in global steps of size dt_max. Note: dt_max must
            # be at least as big as the the biggest step in t_eval (multiplied
            # by some tolerance, here 1.01) to avoid an empty integration window below
            dt_max = self.dt_max
            dt_eval_max = np.max(np.diff(t_eval)) * 1.01
            if dt_max < dt_eval_max:
                pybamm.logger.debug(
                    "Setting dt_max to be as big as the largest step in "
                    f"t_eval ({dt_eval_max})"
                )
                dt_max = dt_eval_max
            termination_due_to_small_dt = False
            first_ts_solved = False
            while t < t_f:
                # Step
                solved = False
                count = 0
                dt = dt_max
                while not solved:
                    # Get window of time to integrate over (so that we return
                    # all the points in t_eval, not just t and t+dt)
                    t_window = np.concatenate(
                        ([t], t_eval[(t_eval > t) & (t_eval < t + dt)])
                    )
                    # Sometimes near events the solver fails between two time
                    # points in t_eval (i.e. no points t < t_i < t+dt for t_i
                    # in t_eval), so we simply integrate from t to t+dt
                    if len(t_window) == 1:
                        t_window = np.array([t, t + dt])

                    if self.mode == "safe":
                        # update integrator with the grid
                        self.create_integrator(model, inputs, t_window)
                    # Try to solve with the current global step, if it fails then
                    # halve the step size and try again.
                    try:
                        pybamm.logger.debug(
                            "Running integrator for "
                            f"{t_window[0]:.2f} < t < {t_window[-1]:.2f}"
                        )
                        current_step_sol = self._run_integrator(
                            model,
                            y0,
                            inputs_dict,
                            inputs,
                            t_window,
                            use_grid=use_grid,
                            extract_sensitivities_in_solution=False,
                        )
                        first_ts_solved = True
                        solved = True
                    except pybamm.SolverError as error:
                        pybamm.logger.debug("Failed, halving step size")
                        dt /= 2
                        count += 1
                        # also reduce maximum step size for future global steps,
                        # but skip them in the beginning
                        # sometimes, for the first integrator smaller timesteps are
                        # needed, but this won't affect the global timesteps. The
                        # global timestep will only be reduced after the first timestep.
                        if first_ts_solved:
                            dt_max = dt
                        if count > self.max_step_decrease_count:
                            message = (
                                "Maximum number of decreased steps occurred at "
                                f"t={t} (final SolverError: '{error}'). "
                                "For a full solution try reducing dt_max (currently, "
                                f"dt_max={dt_max}) and/or reducing the size of the "
                                "time steps or period of the experiment."
                            )
                            if first_ts_solved and self.return_solution_if_failed_early:
                                warnings.warn(
                                    message, pybamm.SolverWarning, stacklevel=2
                                )
                                termination_due_to_small_dt = True
                                break
                            else:
                                raise pybamm.SolverError(
                                    message
                                    + " Set `return_solution_if_failed_early=True` to "
                                    "return the solution object up to the point where "
                                    "failure occured."
                                ) from error
                if termination_due_to_small_dt:
                    break
                # Check if the sign of an event changes, if so find an accurate
                # termination point and exit
                current_step_sol = self._solve_for_event(current_step_sol)
                # assign temporary solve time
                current_step_sol.solve_time = np.nan
                # append solution from the current step to solution
                solution = solution + current_step_sol
                if current_step_sol.termination == "event":
                    break
                else:
                    # update time as time
                    # from which to start the new casadi integrator
                    t = t_window[-1]
                    # update y0 as initial_values
                    # from which to start the new casadi integrator
                    y0 = solution.all_ys[-1][:, -1]

            # now we extract sensitivities from the solution
            if bool(model.calculate_sensitivities):
                solution.sensitivities = True

            solution.check_ys_are_not_too_large()
            return solution

    def _solve_for_event(self, coarse_solution):
        """
        Check if the sign of an event changes, if so find an accurate
        termination point and exit

        Locate the event time using a root finding algorithm and
        event state using interpolation. The solution is then truncated
        so that only the times up to the event are returned
        """
        pybamm.logger.debug("Solving for events")

        model = coarse_solution.all_models[-1]
        inputs_dict = coarse_solution.all_inputs[-1]
        inputs = casadi.vertcat(*[x for x in inputs_dict.values()])

        def find_t_event(sol, typ):
            # Check most recent y to see if any events have been crossed
            if model.terminate_events_eval:
                y_last = sol.all_ys[-1][:, -1]
                crossed_events = np.sign(
                    np.concatenate(
                        [
                            event(sol.t[-1], y_last, inputs)
                            for event in model.terminate_events_eval
                        ]
                    )
                    - 1e-5
                )
            else:
                crossed_events = np.sign([])

            # Return None if no events have been triggered
            if (crossed_events == 1).all():
                return None, None, None

            # get the index of the events that have been crossed
            event_idx = np.where(crossed_events != 1)[0]
            active_events = [model.terminate_events_eval[i] for i in event_idx]

            # loop over events to compute the time at which they were triggered
            t_events = [None] * len(active_events)
            event_idcs_lower = [None] * len(active_events)
            for i, event in enumerate(active_events):
                # Implement our own bisection algorithm for speed
                # This is used to find the time range in which the event is triggered
                # Evaluations of the "event" function are (relatively) expensive
                f_eval = {}

                def f(idx, f_eval=f_eval, event=event):
                    try:
                        return f_eval[idx]
                    except KeyError:
                        # We take away 1e-5 to deal with the case where the event sits
                        # exactly on zero, as can happen when the event switch is used
                        # (fast with events mode)
                        f_eval[idx] = event(sol.t[idx], sol.y[:, idx], inputs) - 1e-5
                        return f_eval[idx]

                def integer_bisect():
                    a_n = 0
                    b_n = len(sol.t) - 1
                    for _ in range(len(sol.t)):
                        if a_n + 1 == b_n:
                            return a_n
                        m_n = (a_n + b_n) // 2
                        f_m_n = f(m_n)
                        if np.isnan(f_m_n):
                            a_n = a_n
                            b_n = m_n
                        elif f_m_n < 0:
                            a_n = a_n
                            b_n = m_n
                        elif f_m_n > 0:
                            a_n = m_n
                            b_n = b_n

                event_idx_lower = integer_bisect()
                if typ == "window":
                    event_idcs_lower[i] = event_idx_lower
                elif typ == "exact":
                    # Linear interpolation between the two indices to find the root time
                    # We could do cubic interpolation here instead but it would be
                    # slower
                    t_lower = sol.t[event_idx_lower]
                    t_upper = sol.t[event_idx_lower + 1]
                    event_lower = abs(f(event_idx_lower))
                    event_upper = abs(f(event_idx_lower + 1))

                    t_events[i] = (event_lower * t_upper + event_upper * t_lower) / (
                        event_lower + event_upper
                    )

            if typ == "window":
                event_idx_lower = np.nanmin(event_idcs_lower)
                return event_idx_lower, None, None
            elif typ == "exact":
                # t_event is the earliest event triggered
                t_event = np.nanmin(t_events)
                # create interpolant to evaluate y in the current integration
                # window
                y_sol = interp1d(sol.t, sol.y, kind="linear")
                y_event = y_sol(t_event)

                closest_event_idx = event_idx[np.nanargmin(t_events)]

                return t_event, y_event, closest_event_idx

        # Find the interval in which the event was triggered
        event_idx_lower, _, _ = find_t_event(coarse_solution, "window")

        # Return the existing solution if no events have been triggered
        if event_idx_lower is None:
            # Flag "final time" for termination
            coarse_solution.termination = "final time"
            return coarse_solution

        # If events have been triggered, we solve for a dense window in the interval
        # where the event was triggered, then find the precise location of the event
        # Solve again with a more dense idx_window, starting from the start of the
        # window where the event was triggered
        t_window_event_dense = np.linspace(
            coarse_solution.t[event_idx_lower],
            coarse_solution.t[event_idx_lower + 1],
            100,
        )

        if self.mode == "safe without grid":
            use_grid = False
        else:
            self.create_integrator(model, inputs, t_window_event_dense)
            use_grid = True

        y0 = coarse_solution.y[:, event_idx_lower]
        dense_step_sol = self._run_integrator(
            model,
            y0,
            inputs_dict,
            inputs,
            t_window_event_dense,
            use_grid=use_grid,
            extract_sensitivities_in_solution=False,
        )

        # Find the exact time at which the event was triggered
        t_event, y_event, closest_event_idx = find_t_event(dense_step_sol, "exact")
        # If this returns None, no event was crossed in dense_step_sol. This can happen
        # if the event crossing was right at the end of the interval in the coarse
        # solution. In this case, return the t and y from the end of the interval
        # (i.e. next point in the coarse solution)
        if y_event is None:  # pragma: no cover
            # This is extremely rare, it's difficult to find a test that triggers this
            # hence no coverage check
            t_event = coarse_solution.t[event_idx_lower + 1]
            y_event = coarse_solution.y[:, event_idx_lower + 1].full().flatten()

        # Return solution truncated at the first coarse event time
        # Also assign t_event
        t_sol = coarse_solution.t[: event_idx_lower + 1]
        y_sol = coarse_solution.y[:, : event_idx_lower + 1]
        solution = pybamm.Solution(
            t_sol,
            y_sol,
            model,
            inputs_dict,
            np.array([t_event]),
            y_event[:, np.newaxis],
            "event",
            all_sensitivities=False,
        )
        solution.integration_time = (
            coarse_solution.integration_time + dense_step_sol.integration_time
        )

        solution.closest_event_idx = closest_event_idx

        return solution

    def create_integrator(self, model, inputs, t_eval=None, use_event_switch=False):
        """
        Method to create a casadi integrator object.
        If t_eval is provided, the integrator uses t_eval to make the grid.
        Otherwise, the integrator has grid [0,1].
        """
        pybamm.logger.debug("Creating CasADi integrator")

        # Use grid if t_eval is given
        use_grid = t_eval is not None
        if use_grid is True:
            t_eval_shifted = t_eval - t_eval[0]
            t_eval_shifted_rounded = np.round(t_eval_shifted, decimals=12).tobytes()
        # Only set up problem once
        if model in self.integrators:
            # If we're not using the grid, we don't need to change the integrator
            if use_grid is False:
                return self.integrators[model]["no grid"]
            # Otherwise, create new integrator with an updated grid
            # We don't need to update the grid if reusing the same t_eval
            # (up to a shift by a constant)
            else:
                if t_eval_shifted_rounded in self.integrators[model]:
                    return self.integrators[model][t_eval_shifted_rounded]
                else:
                    method, problem, options, time_args = self.integrator_specs[model]
                    time_args = [t_eval_shifted[0], t_eval_shifted[1:]]
                    integrator = casadi.integrator(
                        "F", method, problem, *time_args, options
                    )
                    self.integrators[model][t_eval_shifted_rounded] = integrator
                    return integrator
        else:
            rhs = model.casadi_rhs
            algebraic = model.casadi_algebraic

            options = {
                "show_eval_warnings": False,
                **self.extra_options_setup,
                "reltol": self.rtol,
                "abstol": self.atol,
            }

            # set up and solve
            t = casadi.MX.sym("t")
            p = casadi.MX.sym("p", inputs.shape[0])
            y0 = model.y0

            y_diff = casadi.MX.sym("y_diff", rhs(0, y0, p).shape[0])
            y_alg = casadi.MX.sym("y_alg", algebraic(0, y0, p).shape[0])
            y_full = casadi.vertcat(y_diff, y_alg)

            if use_grid is False:
                time_args = []
                # rescale time
                t_min = casadi.MX.sym("t_min")
                t_max = casadi.MX.sym("t_max")
                t_max_minus_t_min = t_max - t_min
                t_scaled = t_min + (t_max - t_min) * t
                # add time limits as inputs
                p_with_tlims = casadi.vertcat(p, t_min, t_max)
            else:
                time_args = [t_eval_shifted[0], t_eval_shifted[1:]]
                # rescale time
                t_min = casadi.MX.sym("t_min")
                # Set dummy parameters for consistency with rescaled time
                t_max_minus_t_min = 1
                t_scaled = t_min + t
                p_with_tlims = casadi.vertcat(p, t_min)

            # define the event switch as the point when an event is crossed
            # we don't do this for ODE models
            # see #1082
            event_switch = 1
            if use_event_switch is True and not algebraic(0, y0, p).is_empty():
                for event in model.casadi_switch_events:
                    event_switch *= event(t_scaled, y_full, p)

            problem = {
                "t": t,
                "x": y_diff,
                # rescale rhs by (t_max - t_min)
                "ode": (t_max_minus_t_min) * rhs(t_scaled, y_full, p) * event_switch,
                "p": p_with_tlims,
            }
            if algebraic(0, y0, p).is_empty():
                method = "cvodes"
            else:
                method = "idas"
                problem.update(
                    {
                        "z": y_alg,
                        "alg": algebraic(t_scaled, y_full, p),
                    }
                )
            integrator = casadi.integrator("F", method, problem, *time_args, options)
            self.integrator_specs[model] = method, problem, options, time_args
            if use_grid is False:
                self.integrators[model] = {"no grid": integrator}
            else:
                self.integrators[model] = {t_eval_shifted_rounded: integrator}

            return integrator

    def _run_integrator(
        self,
        model,
        y0,
        inputs_dict,
        inputs,
        t_eval,
        use_grid=True,
        extract_sensitivities_in_solution=None,
    ):
        """
        Run the integrator.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        y0:
            casadi vector of initial conditions
        inputs_dict : dict, optional
            Any input parameters to pass to the model when solving
        inputs:
            Casadi vector of inputs
        t_eval : numeric type
            The times at which to compute the solution
        use_grid: bool, optional
            Determines whether the casadi solver uses a grid or rescales time to (0,1)
        extract_sensitivities_in_solution: bool or None
            If None, then the sensitivities are extracted within the
            :class:`pybamm.Solution` object returned, only if present in the solution.
            Setting to True or False will override this behaviour, forcing the
            sensitivities to be extracted or not (it is up to the caller to determine if
            the sensitivities are in fact present)
        """

        pybamm.logger.debug("Running CasADi integrator")

        # are we solving explicit forward equations?
        explicit_sensitivities = bool(model.calculate_sensitivities)
        # by default we extract sensitivities in the solution if we
        # are calculating the sensitivities
        if extract_sensitivities_in_solution is None:
            extract_sensitivities_in_solution = explicit_sensitivities

        if use_grid is True:
            pybamm.logger.spam("Calculating t_eval_shifted")
            t_eval_shifted = t_eval - t_eval[0]
            t_eval_shifted_rounded = np.round(t_eval_shifted, decimals=12).tobytes()
            pybamm.logger.spam("Finished calculating t_eval_shifted")
            integrator = self.integrators[model][t_eval_shifted_rounded]
        else:
            integrator = self.integrators[model]["no grid"]

        len_rhs = model.concatenated_rhs.size
        len_alg = model.concatenated_algebraic.size

        # Check y0 to see if it includes sensitivities
        if explicit_sensitivities:
            num_parameters = model.len_rhs_sens // model.len_rhs
            len_rhs = len_rhs * (num_parameters + 1)
            len_alg = len_alg * (num_parameters + 1)

        y0_diff = y0[:len_rhs]
        y0_alg_exact = y0[len_rhs:]
        if self.perturb_algebraic_initial_conditions and len_alg > 0:
            # Add a tiny perturbation to the algebraic initial conditions
            # For some reason this helps with convergence
            # The actual value of the initial conditions for the algebraic variables
            # doesn't matter
            y0_alg = y0_alg_exact * (1 + 1e-6 * casadi.DM(np.random.rand(len_alg)))
        else:
            y0_alg = y0_alg_exact
        pybamm.logger.spam("Finished preliminary setup for integrator run")

        # Solve
        # Try solving
        if use_grid is True:
            t_min = t_eval[0]
            inputs_with_tmin = casadi.vertcat(inputs, t_min)
            # Call the integrator once, with the grid
            timer = pybamm.Timer()
            pybamm.logger.debug("Calling casadi integrator")
            try:
                casadi_sol = integrator(
                    x0=y0_diff, z0=y0_alg, p=inputs_with_tmin, **self.extra_options_call
                )
            except RuntimeError as error:
                # If it doesn't work raise error
                pybamm.logger.debug(f"Casadi integrator failed with error {error}")
                raise pybamm.SolverError(error.args[0]) from error
            pybamm.logger.debug("Finished casadi integrator")
            integration_time = timer.time()
            # Manually add initial conditions and concatenate
            x_sol = casadi.horzcat(y0_diff, casadi_sol["xf"])
            if len_alg > 0:
                z_sol = casadi.horzcat(y0_alg_exact, casadi_sol["zf"])
                y_sol = casadi.vertcat(x_sol, z_sol)
            else:
                y_sol = x_sol
            sol = pybamm.Solution(
                t_eval,
                y_sol,
                model,
                inputs_dict,
                all_sensitivities=extract_sensitivities_in_solution,
                check_solution=False,
            )
            sol.integration_time = integration_time
            return sol
        else:
            # Repeated calls to the integrator
            x = y0_diff
            z = y0_alg_exact
            y_diff = x
            y_alg = z
            for i in range(len(t_eval) - 1):
                t_min = t_eval[i]
                t_max = t_eval[i + 1]
                inputs_with_tlims = casadi.vertcat(inputs, t_min, t_max)
                timer = pybamm.Timer()
                try:
                    casadi_sol = integrator(
                        x0=x, z0=z, p=inputs_with_tlims, **self.extra_options_call
                    )
                except RuntimeError as error:
                    # If it doesn't work raise error
                    pybamm.logger.debug(f"Casadi integrator failed with error {error}")
                    raise pybamm.SolverError(error.args[0]) from error
                integration_time = timer.time()
                x = casadi_sol["xf"]
                z = casadi_sol["zf"]
                y_diff = casadi.horzcat(y_diff, x)
                if not z.is_empty():
                    y_alg = casadi.horzcat(y_alg, z)
            if z.is_empty():
                y_sol = y_diff
            else:
                y_sol = casadi.vertcat(y_diff, y_alg)

            sol = pybamm.Solution(
                t_eval,
                y_sol,
                model,
                inputs_dict,
                all_sensitivities=extract_sensitivities_in_solution,
                check_solution=False,
            )
            sol.integration_time = integration_time
            return sol
